import json
import unittest

import responses
from six import iteritems

import pyfacebook


class ApiStoriesTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram/apidata/stories/"
    BASE_URL = "https://graph.facebook.com/{}/".format(pyfacebook.Api.VALID_API_VERSIONS[-1])

    with open(BASE_PATH + "stories_default.json", "rb") as f:
        STORIES_DEFAULT = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "stories_fields_p1.json", "rb") as f:
        STORIES_FIELDS_p1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "stories_fields_p2.json", "rb") as f:
        STORIES_FIELDS_p2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "story_info.json", "rb") as f:
        STORY_INFO = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "stories_info.json", "rb") as f:
        STORIES_INFO = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.instagram_business_id = "17841406338772941"
        self.api = pyfacebook.IgProApi(
            app_id="123456", app_secret="secret",
            long_term_token="token"
        )

    def testGetStories(self):
        # test count
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/stories", json=self.STORIES_DEFAULT)

            res = self.api.get_user_stories(
                user_id=self.instagram_business_id,
                count=2,
                return_json=True
            )
            self.assertEqual(len(res), 2)

        # test all items
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/stories", json=self.STORIES_FIELDS_p1)
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/stories", json=self.STORIES_FIELDS_p2)

            res = self.api.get_user_stories(
                user_id=self.instagram_business_id,
                fields=["id", "media_type", "media_url", "username"],
                count=None,
            )
            self.assertEqual(len(res), 3)

    def testGetStoryInfo(self):
        story_id = "17876166859931045"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + story_id, json=self.STORY_INFO)

            story = self.api.get_story_info(
                story_id=story_id,
            )
            self.assertEqual(story.caption, "Nice")
            self.assertEqual(story.id, story_id)

            story_json = self.api.get_story_info(
                story_id=story_id,
                return_json=True
            )
            self.assertEqual(story_json["id"], story_id)

    def testGetStoriesInfo(self):
        story_ids = ["17876166859931045", "18168273289049440"]
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL, json=self.STORIES_INFO)

            stories = self.api.get_stories_info(
                story_ids=story_ids,
            )
            for _id, data in iteritems(stories):
                self.assertIn(_id, story_ids)
                self.assertIn(_id, data.id)

            stories_json = self.api.get_stories_info(
                story_ids=story_ids,
                return_json=True
            )
            for _id, data in iteritems(stories_json):
                self.assertIn(_id, story_ids)
                self.assertIn(_id, data["id"])



from attr import attrs, attrib, asdict
from typing import Optional

__all__ = [
    "PyFacebookError", "ErrorCode", "ErrorMessage",
    "PyFacebookException", "PyFacebookDeprecationWaring"
]


class PyFacebookError(Exception):
    """ Base class for PyFacebook errors"""

    @property
    def message(self):
        """ return the error's first arg """
        return self.args[0]  # pragma: no cover


class ErrorCode:
    HTTP_ERROR = 10000
    MISSING_PARAMS = 10001
    INVALID_PARAMS = 10002
    NOT_SUPPORT_METHOD = 10010


@attrs
class ErrorMessage(object):
    code = attrib(default=None, type=Optional[str])
    message = attrib(default=None, type=Optional[str])
    error_type = attrib(default="PyFacebookException", type=Optional[str])


class PyFacebookException(Exception):
    """
    Library exception class.

    Show internal error and api response error.
    """

    def __init__(self, data):
        self.data = None
        self.message = None
        self.type = None
        self.code = None
        self.fbtrace_id = None
        self.error_type = None
        self.initial(data)

    def initial(self, data):
        """
        Args:
             data (dict,ErrorMessage):
                1. Internal error data is an instance of ErrorMessage.
                2. Api response error is dict.
                   Refer: https://developers.facebook.com/docs/graph-api/using-graph-api/error-handling
        """
        if isinstance(data, ErrorMessage):
            self.message = data.message
            self.error_type = data.error_type
            self.type = data.error_type
            self.code = data.code
            self.data = asdict(data)
        elif isinstance(data, dict):
            self.error_type = "FacebookException"
            self.message = data.get("message")
            self.type = data.get("type")
            self.code = data.get("code")
            self.fbtrace_id = data.get("fbtrace_id")
            self.data = {"error": data}

    def __repr__(self):
        return (
            "{0}(code={1},type={2},message={3})".format(
                self.error_type, self.code, self.type, self.message
            )
        )

    def __str__(self):
        return self.__repr__()


class PyFacebookDeprecationWaring(DeprecationWarning):
    pass



"""
    function's to validate parameters.
"""
import logging

from typing import Optional, Union

import pyfacebook

logger = logging.getLogger(__name__)


def enf_comma_separated(
        field,  # type: str
        value,  # type: Optional[Union[str, list, tuple, set]]
):
    # type: (...) -> str
    """
    Check to see if field's value type belong to correct type.
    If it is, return api need value, otherwise, raise a PyYouTubeException.

    Args:
        field (str):
            Name of the field you want to do check.
        value (str, list, tuple, set, Optional)
            Value for the field.

    Returns:
        Api needed string
    """
    try:
        # if value point with string, not check.
        if isinstance(value, str):
            value = [value]
        elif isinstance(value, (list, tuple)):
            pass
        elif isinstance(value, set):
            logger.warning("Note: The order of the set is unreliable.")
            pass
        else:
            raise pyfacebook.PyFacebookError({
                "message": "Parameter ({0}) must be single str,comma-separated str,list,tuple or set".format(field),
            })

        # loop value list and remove repeat item
        seen = {}
        res = []
        for item in value:
            if item in seen:
                continue
            seen[item] = 1
            res.append(item)
        return ",".join(res)
    except (TypeError, ValueError):
        raise pyfacebook.PyFacebookError({
            "message": "Parameter ({0}) must be single str,comma-separated str,list,tuple or set".format(field),
        })



import json
import unittest

from pyfacebook.error import ErrorCode, ErrorMessage, PyFacebookException


class ErrorTest(unittest.TestCase):
    BASE_PATH = "testdata/"
    with open(BASE_PATH + "error.json", "rb") as f:
        ERROR_DATA = json.loads(f.read().decode("utf-8"))

    def testResponseError(self):
        ex = PyFacebookException(self.ERROR_DATA["error"])

        self.assertEqual(ex.code, 100)
        self.assertEqual(ex.type, "OAuthException")
        self.assertEqual(ex.error_type, "FacebookException")
        error_msg = (
            "FacebookException(code=100,type=OAuthException,message=(#100) "
            "Pages Public Content Access requires either app secret proof or an app token)"
        )
        self.assertEqual(repr(ex), error_msg)
        self.assertTrue(str(ex), error_msg)

    def testErrorMessage(self):
        error = ErrorMessage(code=ErrorCode.HTTP_ERROR, message="error")

        ex = PyFacebookException(error)

        self.assertEqual(ex.code, 10000)
        self.assertEqual(ex.message, "error")
        self.assertEqual(ex.error_type, "PyFacebookException")



import json
import unittest

import responses

import pyfacebook


class ApiUserTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram/apidata/users/"
    BASE_URL = "https://graph.facebook.com/{}/".format(pyfacebook.Api.VALID_API_VERSIONS[-1])

    with open(BASE_PATH + "user_default_fields.json", "rb") as f:
        USER_DEFAULT = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "user_fields.json", "rb") as f:
        USER_FIELDS = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.instagram_business_id = "17841406338772941"
        self.api = pyfacebook.IgProApi(
            app_id="123456", app_secret="secret",
            long_term_token="token"
        )

    def testGetUserInfo(self):
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id, json=self.USER_DEFAULT)
            m.add("GET", self.BASE_URL + self.instagram_business_id, json=self.USER_FIELDS)

            user_full_fields = self.api.get_user_info(
                user_id=self.instagram_business_id
            )
            self.assertEqual(user_full_fields.id, self.instagram_business_id)
            self.assertEqual(user_full_fields.follows_count, 17)
            self.assertEqual(user_full_fields.username, "ikroskun")

            user_fields = self.api.get_user_info(
                user_id=self.instagram_business_id,
                fields=["id", "username", "name"],
                return_json=True
            )
            self.assertEqual(user_fields["id"], self.instagram_business_id)
            self.assertEqual(user_fields["name"], "LiuKun")



import json
import unittest

import responses

import pyfacebook


class ApiUserTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram_basic/apidata/user/"

    with open(BASE_PATH + "user_info_default.json", "rb") as f:
        USER_INFO_DEFAULT = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "user_with_fields.json", "rb") as f:
        USER_INFO_FIELDS = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.api = pyfacebook.IgBasicApi(long_term_token="token")

    def testGetUserInfo(self):
        user_id = "17841406338772941"

        with responses.RequestsMock() as m:
            m.add("GET", self.api.GRAPH_URL + "me", json=self.USER_INFO_DEFAULT)
            m.add("GET", self.api.GRAPH_URL + user_id, json=self.USER_INFO_FIELDS)

            me_info = self.api.get_user_info()
            self.assertEqual(me_info.id, user_id)
            self.assertEqual(me_info.account_type, "BUSINESS")

            user_info = self.api.get_user_info(
                user_id=user_id,
                fields=["id", "username"],
                return_json=True
            )
            self.assertEqual(user_info["id"], user_id)
            self.assertEqual(user_info["username"], "ikroskun")



"""
    These are models for video entity.

    Refer: https://developers.facebook.com/docs/graph-api/reference/video
"""

from attr import attrs, attrib
from typing import List, Optional

from .base import BaseModel
from .common import Privacy, Place
from .mixins import LikesAndCommentsSummaryField
from .._compat import str


@attrs
class VideoFormat(BaseModel):
    embed_html = attrib(default=None, type=Optional[str], repr=False)
    filter = attrib(default=None, type=Optional[str], repr=False)
    height = attrib(default=None, type=Optional[int], repr=False)
    picture = attrib(default=None, type=Optional[str])
    width = attrib(default=None, type=Optional[int], repr=False)


# TODO
# @attrs
# class VideoFrom(BaseModel):
#     id = attrib(default=None, type=Optional[str])
#     name = attrib(default=None, type=Optional[str])

@attrs
class VideoPlace(Place):
    pass


@attrs
class VideoStatus(BaseModel):
    processing_progress = attrib(default=None, type=Optional[str], repr=False)
    video_status = attrib(default=None, type=Optional[str])


@attrs
class Video(BaseModel, LikesAndCommentsSummaryField):
    """
    A class representing the video info.
    """

    id = attrib(default=None, type=Optional[str])
    ad_breaks = attrib(default=None, type=Optional[List[int]], repr=False)
    backdated_time = attrib(default=None, type=Optional[str], repr=False)
    backdated_time_granularity = attrib(default=None, type=Optional[str], repr=False)
    content_category = attrib(default=None, type=Optional[str], repr=False)
    content_tags = attrib(default=None, type=Optional[List[str]], repr=False)
    created_time = attrib(default=None, type=Optional[str])
    custom_labels = attrib(default=None, type=Optional[List[str]], repr=False)
    description = attrib(default=None, type=Optional[str])
    embed_html = attrib(default=None, type=Optional[str], repr=False)
    embeddable = attrib(default=None, type=Optional[bool], repr=False)
    format = attrib(default=None, type=Optional[List[VideoFormat]], repr=False)
    # TODO now cattr not allow rename
    # from
    icon = attrib(default=None, type=Optional[str], repr=False)
    is_crosspost_video = attrib(default=None, type=Optional[bool], repr=False)
    is_crossposting_eligible = attrib(default=None, type=Optional[bool], repr=False)
    is_instagram_eligible = attrib(default=None, type=Optional[bool], repr=False)
    length = attrib(default=None, type=Optional[float], repr=False)
    live_status = attrib(default=None, type=Optional[str], repr=False)
    permalink_url = attrib(default=None, type=Optional[str], repr=False)
    place = attrib(default=None, type=Optional[VideoPlace], repr=False)
    premiere_living_room_status = attrib(default=None, type=Optional[str], repr=False)
    privacy = attrib(default=None, type=Optional[Privacy], repr=False)
    published = attrib(default=None, type=Optional[bool], repr=False)
    scheduled_publish_time = attrib(default=None, type=Optional[str], repr=False)
    source = attrib(default=None, type=Optional[str], repr=False)
    status = attrib(default=None, type=Optional[VideoStatus], repr=False)
    title = attrib(default=None, type=Optional[str], repr=False)
    universal_video_id = attrib(default=None, type=Optional[str], repr=False)
    updated_time = attrib(default=None, type=Optional[str], repr=False)


@attrs
class VideoCaption(BaseModel):
    """
    A class representing the video caption info.
    """

    create_time = attrib(default=None, type=Optional[str])
    is_auto_generated = attrib(default=None, type=Optional[bool], repr=False)
    is_default = attrib(default=None, type=Optional[bool], repr=False)
    locale = attrib(default=None, type=Optional[str], repr=False)
    locale_name = attrib(default=None, type=Optional[str], repr=False)
    uri = attrib(default=None, type=Optional[str], repr=False)



import json
import unittest

import pyfacebook.models as models


class LiveVideoModelTest(unittest.TestCase):
    BASE_PATH = "testdata/facebook/models/live/"

    with open(BASE_PATH + 'live_video.json', 'rb') as f:
        LIVE_VIDEO_INFO = json.loads(f.read().decode('utf-8'))

    def testLiveVideo(self):
        m = models.LiveVideo.new_from_json_dict(self.LIVE_VIDEO_INFO)

        self.assertEqual(m.id, "178598660650827")
        self.assertEqual(m.status, "LIVE")
        self.assertEqual(m.ingest_streams[0].stream_id, "0")
        self.assertEqual(m.ingest_streams[0].stream_health.video_bitrate, 2511983)
        self.assertEqual(m.comments.total_count, 45)



import json
import unittest

import responses
from six import iteritems

import pyfacebook


class PostApiTest(unittest.TestCase):
    BASE_PATH = "testdata/facebook/apidata/posts/"
    BASE_URL = "https://graph.facebook.com/v8.0/"
    PAGE_ID = "2121008874780932"

    with open(BASE_PATH + "feeds_default_fields_p1.json", "rb") as f:
        FEED_PAGED_1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "feeds_default_fields_p2.json", "rb") as f:
        FEED_PAGED_2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "posts_data.json", "rb") as f:
        POSTS_DATA = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "published_posts.json", "rb") as f:
        PUBLISHED_POSTS_DATA = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "tagged_posts.json", "rb") as f:
        TAGGED_POSTS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "single_post_data.json", "rb") as f:
        SINGLE_POST_DATA = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "single_post_data_fields.json", "rb") as f:
        SINGLE_POST_DATA_FIELDS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "multi_posts_data.json", "rb") as f:
        MULTI_POST_DATA = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "multi_posts_data_fields.json", "rb") as f:
        MULTI_POST_DATA_FIELDS = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.api = pyfacebook.Api(
            app_id="12345678",
            app_secret="secret",
            long_term_token="token",
            version="v8.0"
        )

    def testPageFeed(self):
        # test base count
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.PAGE_ID + "/feed", json=self.FEED_PAGED_1)
            m.add("GET", self.BASE_URL + self.PAGE_ID + "/feed", json=self.FEED_PAGED_2)

            posts = self.api.get_page_feeds(
                page_id=self.PAGE_ID,
                count=12,
                access_token="token",
                return_json=True
            )

            self.assertEqual(len(posts), 12)

        # test count is None
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.PAGE_ID + "/feed", json=self.FEED_PAGED_1)
            m.add("GET", self.BASE_URL + self.PAGE_ID + "/feed", json=self.FEED_PAGED_2)

            posts = self.api.get_page_feeds(
                page_id=self.PAGE_ID,
                count=None,
                access_token="token",
            )
            self.assertEqual(len(posts), 15)

    def testPagePosts(self):
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.PAGE_ID + "/posts", json=self.POSTS_DATA)

            posts = self.api.get_page_posts(
                page_id=self.PAGE_ID,
                count=None,
                fields=["id", "message", "created_time", "updated_time"],
                limit=20
            )

            self.assertEqual(len(posts), 15)

    def testPublishedPosts(self):
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.PAGE_ID + "/published_posts", json=self.POSTS_DATA)

            posts = self.api.get_page_published_posts(
                page_id=self.PAGE_ID,
                fields="id,message,created_time,updated_time",
                count=None,
                limit=20,
                access_token="my page token",
            )
            self.assertEqual(len(posts), 15)

    def testTaggedPosts(self):
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.PAGE_ID + "/tagged", json=self.TAGGED_POSTS)

            posts = self.api.get_page_tagged_posts(
                page_id=self.PAGE_ID,
                fields="id,message,created_time,updated_time",
                count=None,
                limit=20,
                access_token="my page token",
            )
            self.assertEqual(len(posts), 3)

    def testGetPostInfo(self):
        post_id = "175154750010052_424924701699721"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + post_id, json=self.SINGLE_POST_DATA)

            post = self.api.get_post_info(
                post_id=post_id,
                return_json=True
            )
            self.assertEqual(post["id"], post_id)
        # test provide fields
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + post_id, json=self.SINGLE_POST_DATA_FIELDS)

            post = self.api.get_post_info(
                post_id=post_id,
                fields=("id", "message", "created_time", "updated_time")
            )
            self.assertEqual(post.id, post_id)

    def testGetPosts(self):
        ids = ["2121008874780932_2427414477473702", "175154750010052_424924701699721"]
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL, json=self.MULTI_POST_DATA)

            posts = self.api.get_posts(
                ids=ids,
                return_json=True
            )
            for _id, data in iteritems(posts):
                self.assertIn(_id, ids)
                self.assertEqual(_id, data["id"])

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL, json=self.MULTI_POST_DATA_FIELDS)

            posts = self.api.get_posts(
                ids=ids,
                fields="id,message,created_time,updated_time",
            )
            for _id, data in iteritems(posts):
                self.assertIn(_id, ids)
                self.assertEqual(_id, data.id)



import json
import unittest

import pyfacebook.models as models


class PostModelTest(unittest.TestCase):
    BASE_PATH = "testdata/facebook/models/posts/"

    with open(BASE_PATH + 'post.json', 'rb') as f:
        POST_INFO = json.loads(f.read().decode('utf-8'))

    def testPost(self):
        m = models.Post.new_from_json_dict(self.POST_INFO)

        self.assertEqual(m.id, "565225540184937_4018908568149933")
        self.assertEqual(m.created_time, "2020-02-18T05:20:33+0000")
        self.assertEqual(len(m.attachments), 1)
        self.assertEqual(m.attachments[0].media_type, "album")
        self.assertEqual(len(m.attachments[0].subattachments), 7)
        self.assertEqual(m.attachments[0].subattachments[0].media.image.height, 405)
        self.assertTrue(m.attachments[0].target.url)
        self.assertEqual(m.comments.total_count, 47)
        self.assertEqual(m.reactions.total_count, 8878)
        self.assertEqual(m.reactions.viewer_reaction, "NONE")
        self.assertEqual(m.like.total_count, 8575)



"""
    Base Api Impl
"""
import hashlib
import hmac
import logging
import re
import time
import six
import warnings
from typing import Dict, Optional, Union, List

import requests
from requests import Response
from requests_oauthlib import OAuth2Session
from requests_oauthlib.compliance_fixes.facebook import facebook_compliance_fix

from pyfacebook.error import PyFacebookException, ErrorMessage, ErrorCode, PyFacebookDeprecationWaring
from pyfacebook.models import AccessToken, AuthAccessToken
from pyfacebook.ratelimit import RateLimit, PercentSecond


class BaseApi(object):
    VALID_API_VERSIONS = ["v4.0", "v5.0", "v6.0", "v7.0", "v8.0", "v9.0", "v10.0", "v11.0"]
    GRAPH_URL = "https://graph.facebook.com/"
    DEFAULT_AUTHORIZATION_URL = 'https://www.facebook.com/dialog/oauth'
    DEFAULT_EXCHANGE_ACCESS_TOKEN_URL = 'https://graph.facebook.com/oauth/access_token'
    DEFAULT_REDIRECT_URI = 'https://localhost/'
    DEFAULT_SCOPE = []
    DEFAULT_STATE = 'PyFacebook'

    def __init__(self,
                 app_id=None,  # type: Optional[str]
                 app_secret=None,  # type: Optional[str]
                 short_token=None,  # type: Optional[str]
                 long_term_token=None,  # type: Optional[str]
                 application_only_auth=False,  # type: bool
                 initial_access_token=True,  # type: bool
                 version=None,  # type: Optional[str]
                 timeout=None,  # type: Optional[int]
                 sleep_on_rate_limit=False,  # type: bool
                 sleep_seconds_mapping=None,  # type: Dict[int, int]
                 proxies=None,  # type: Optional[dict]
                 debug_http=False  # type: bool
                 ):
        # type: (...) -> None
        """
        :param app_id: Your app id.
        :param app_secret: Your app secret.
        :param short_token: short-lived token
        :param long_term_token: long-lived token.
        :param application_only_auth: Use the `App Access Token` only.
        :param initial_access_token: If you want use api do authorize, set this with False.
        :param version: The version for the graph api.
        :param timeout: Request time out
        :param sleep_on_rate_limit: Use this will sleep between two request.
        :param sleep_seconds_mapping: Mapping for percent to sleep.
        :param proxies: Your proxies config.
        :param debug_http: Set to True to enable debug output from urllib when performing
        any HTTP requests.  Defaults to False.
        """
        self.app_id = app_id
        self.app_secret = app_secret
        self.short_token = short_token
        self.__timeout = timeout
        self.base_url = self.GRAPH_URL
        self.proxies = proxies
        self.session = requests.Session()
        self.sleep_on_rate_limit = sleep_on_rate_limit
        self.instagram_business_id = None
        self._debug_http = debug_http
        self.authorization_url = self.DEFAULT_AUTHORIZATION_URL
        self.exchange_access_token_url = self.DEFAULT_EXCHANGE_ACCESS_TOKEN_URL
        self.redirect_uri = self.DEFAULT_REDIRECT_URI
        self.scope = self.DEFAULT_SCOPE
        self.auth_session = None  # Authorization session
        self.rate_limit = RateLimit()

        if version is None:
            # default version is last new.
            self.version = self.VALID_API_VERSIONS[-1]
        else:
            version = str(version)
            if not version.startswith('v'):
                version = 'v' + version
            version_regex = re.compile(r"^v\d*.\d{1,2}$")
            match = version_regex.search(str(version))
            if match is not None:
                if version not in self.VALID_API_VERSIONS:
                    raise PyFacebookException(ErrorMessage(
                        code=ErrorCode.INVALID_PARAMS,
                        message="Valid API version are {}".format(",".join(self.VALID_API_VERSIONS))
                    ))
                else:
                    self.version = version
            else:
                raise PyFacebookException(ErrorMessage(
                    code=ErrorCode.INVALID_PARAMS,
                    message="Version string is invalid for {0}. You can provide with like: 5.0 or v5.0".format(version),
                ))

        self.sleep_seconds_mapping = self._build_sleep_seconds_resource(sleep_seconds_mapping)

        if long_term_token:
            self._access_token = long_term_token
        elif short_token and all([self.app_id, self.app_secret]):
            token = self.get_long_token(app_id=self.app_id, app_secret=self.app_secret, short_token=self.short_token)
            self._access_token = token.access_token
        elif application_only_auth and all([self.app_id, app_secret]):
            token = self.get_app_token()
            self._access_token = token.access_token
        elif not initial_access_token and all([self.app_id, app_secret]):
            self._access_token = None
        else:
            raise PyFacebookException(ErrorMessage(
                code=ErrorCode.MISSING_PARAMS,
                message=(
                    "You can initial api with three methods: \n"
                    "1. Just provide long(short) lived token or app access token with param `long_term_token`.\n"
                    "2. Provide a short lived token and app credentials. Api will auto exchange long term token.\n"
                    "3. Provide app credentials and with application_only_auth set to true. "
                    "Api will auto get and use app access token.\n"
                    "4. Provide app credentials and prepare for do authorize(This will not retrieve access token)"
                )
            ))

        if debug_http:
            from six.moves import http_client
            http_client.HTTPConnection.debuglevel = 1

            logging.basicConfig()  # you need to initialize logging, otherwise you will not see anything from requests
            logging.getLogger().setLevel(logging.DEBUG)
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True

    @staticmethod
    def _build_sleep_seconds_resource(sleep_seconds_mapping):
        # type: (Optional[Dict]) -> Optional[List[PercentSecond]]
        """
        Sort and convert data
        :param sleep_seconds_mapping: mapping for sleep.
        :return:
        """
        if sleep_seconds_mapping is None:
            return None
        mapping_list = [PercentSecond(percent=p, seconds=s) for p, s in six.iteritems(sleep_seconds_mapping)]
        return sorted(mapping_list, key=lambda ps: ps.percent)

    @staticmethod
    def _generate_secret_proof(secret, access_token):
        # type: (str, str) -> Optional[str]
        if secret is None:
            logging.warning(
                "Calls from a server can be better secured by adding a parameter called appsecret_proof. "
                "And need your app secret."
            )
            return None
        return hmac.new(secret.encode("utf-8"), msg=access_token.encode("utf-8"), digestmod=hashlib.sha256).hexdigest()

    def _request(self, path, method="GET", args=None, post_args=None, enforce_auth=True):
        # type: (str, str, Optional[dict], Optional[dict], bool) -> Response
        """
        Build the request and send request to Facebook.
        :param path: The path for resource on facebook.
        :param method: Http methods.
        :param args: GET parameters.
        :param post_args: POST parameters.
        :param enforce_auth: Set to True mean this request need access token.
        :return: The Response instance.
        """
        if args is None:
            args = dict()
        if post_args is not None:
            method = "POST"
        if enforce_auth:
            if method == "POST":
                post_args = {}
            if method == "POST" and "access_token" not in post_args:
                post_args["access_token"] = self._access_token
            elif method == "GET" and "access_token" not in args:
                args["access_token"] = self._access_token

            # add appsecret_proof parameter
            # Refer: https://developers.facebook.com/docs/graph-api/securing-requests/
            if method == "POST" and "appsecret_proof" not in post_args:
                secret_proof = self._generate_secret_proof(self.app_secret, post_args["access_token"])
                if secret_proof is not None:
                    post_args["appsecret_proof"] = secret_proof
            elif method == "GET" and "appsecret_proof" not in args:
                secret_proof = self._generate_secret_proof(self.app_secret, args["access_token"])
                if secret_proof is not None:
                    args["appsecret_proof"] = secret_proof

        # check path
        if not path.startswith("https"):
            path = self.base_url + path
        try:
            response = self.session.request(
                method,
                path,
                timeout=self.__timeout,
                params=args,
                data=post_args,
                proxies=self.proxies
            )
        except requests.HTTPError as e:
            raise PyFacebookException(ErrorMessage(code=ErrorCode.HTTP_ERROR, message=e.args[0]))
        headers = response.headers
        self.rate_limit.set_limit(headers)
        if self.sleep_on_rate_limit:
            sleep_seconds = self.rate_limit.get_sleep_seconds(sleep_data=self.sleep_seconds_mapping)
            time.sleep(sleep_seconds)
        return response

    def _parse_response(self, response):
        # type: (Response) -> dict
        data = response.json()
        self._check_graph_error(data)
        return data

    @staticmethod
    def _check_graph_error(data):
        # type: (dict) -> None
        """
        Check the facebook response data. If have error raise a PyFacebookException.

        :param data: The data return by facebook.
        :return: None
        """
        if 'error' in data:
            error_data = data['error']
            raise PyFacebookException(error_data)

    def get_long_token(self, short_token, app_id=None, app_secret=None, return_json=False):
        # type: (str, Optional[str], Optional[str], bool) -> Optional[Union[AuthAccessToken, Dict]]
        """
        Generate a long-lived token from a short-lived User access token.
        A long-lived token generally lasts about 60 days.
        :param app_id: Your app id.
        :param app_secret: Your app secret.
        :param short_token: short-lived token from the app.
        :param return_json: Set to false will return instance of AuthAccessToken.
        Or return json data. Default is false.
        """
        if app_id is None:
            app_id = self.app_id
            app_secret = self.app_secret

        response = self._request(
            method='GET',
            path='{}/oauth/access_token'.format(self.version),
            args={
                'grant_type': 'fb_exchange_token',
                'client_id': app_id,
                'client_secret': app_secret,
                'fb_exchange_token': short_token
            },
            enforce_auth=False
        )
        data = self._parse_response(response)
        if return_json:
            return data
        else:
            return AuthAccessToken.new_from_json_dict(data)

    def get_app_token(self, return_json=False):
        # type: (bool)-> Optional[Union[Dict, AuthAccessToken]]
        """
        Use app credentials to generate an app access token.
        :param return_json: Set to false will return instance of AuthAccessToken.
        Or return json data. Default is false.
        """
        resp = self._request(
            method="GET",
            path="{}/oauth/access_token".format(self.version),
            args={
                'grant_type': 'client_credentials',
                'client_id': self.app_id,
                'client_secret': self.app_secret,
            },
            enforce_auth=False
        )
        data = self._parse_response(resp)
        if return_json:
            return data
        else:
            return AuthAccessToken.new_from_json_dict(data)

    def get_token_info(self, input_token=None, return_json=False):
        # type: (Optional[str], bool) -> Optional[Union[Dict, AccessToken]]
        """
        Obtain the current access token info if provide the app_id and app_secret.
        :param input_token:
        :param return_json: Set to false will return instance of AccessToken.
        Or return json data. Default is false.
        """
        if input_token is None:
            input_token = self._access_token

        if all([self.app_id, self.app_secret]):
            access_token = "{0}|{1}".format(self.app_id, self.app_secret)
        else:
            access_token = self._access_token

        args = {
            "input_token": input_token,
            "access_token": access_token,
        }
        resp = self._request(
            '{0}/debug_token'.format(self.version),
            args=args
        )
        data = self._parse_response(resp)

        if return_json:
            return data["data"]
        else:
            return AccessToken.new_from_json_dict(data['data'])

    def _get_oauth_session(self, redirect_uri=None, scope=None, **kwargs):
        """
        build session for oauth flow.
        :param redirect_uri: The URL that you want to redirect the person logging in back to.
        :param scope: A list of Permissions to request from the person using your app.
        :param kwargs: Extend args for oauth.
        :return:
        """
        if redirect_uri is None:
            redirect_uri = self.redirect_uri

        if scope is None:
            scope = self.scope
        session = OAuth2Session(
            client_id=self.app_id, scope=scope, redirect_uri=redirect_uri,
            state=self.DEFAULT_STATE, **kwargs
        )
        session = facebook_compliance_fix(session)
        return session

    def get_authorization_url(self, redirect_uri=None, scope=None, **kwargs):
        # type: (str, List, Dict) -> (str, str)
        """
        Build authorization url to do authorize.

        Refer: https://developers.facebook.com/docs/facebook-login/manually-build-a-login-flow

        :param redirect_uri: The URL that you want to redirect the person logging in back to.
        Note: Your redirect uri need be set to `Valid OAuth redirect URIs` items in App Dashboard.
        :param scope: A list of Permissions to request from the person using your app.
        :param kwargs: Extend args for oauth.
        :return: Authorization url and state.
        """
        if not all([self.app_id, self.app_secret]):
            raise PyFacebookException(ErrorMessage(
                code=ErrorCode.MISSING_PARAMS,
                message="To do authorization need your app credentials"
            ))

        session = self._get_oauth_session(redirect_uri=redirect_uri, scope=scope, **kwargs)

        authorization_url, state = session.authorization_url(
            url=self.authorization_url
        )

        return authorization_url, state

    def exchange_access_token(self, response, redirect_uri=None, return_json=False):
        # type: (str, str, bool) -> Union[AuthAccessToken, Dict]
        """
        :param response: The whole response url for your previous authorize step.
        :param redirect_uri: The URL that you want to redirect the person logging in back to.
        :param return_json: Set to false will return instance of AuthAccessToken.
        Or return json data. Default is false.
        :return:
        """
        session = self._get_oauth_session(redirect_uri=redirect_uri)

        session.fetch_token(
            self.exchange_access_token_url, client_secret=self.app_secret,
            authorization_response=response
        )

        self._access_token = session.access_token

        if return_json:
            return session.token
        else:
            return AuthAccessToken.new_from_json_dict(session.token)

    def exchange_page_token(self, page_id, access_token=None):
        # type: (str, Optional[str]) -> str
        """
        Use user access token to exchange page(managed by that user) access token.

        Refer:
            1. https://developers.facebook.com/docs/pages/access-tokens
            2. https://developers.facebook.com/docs/facebook-login/access-tokens

        :param page_id: The id for page
        :param access_token: user access token
        """
        if access_token is None:
            access_token = self._access_token

        args = {
            "access_token": access_token,
            "fields": "access_token"
        }
        resp = self._request(
            method="GET",
            path="{version}/{page_id}".format(version=self.version, page_id=page_id),
            args=args,
            enforce_auth=False
        )

        data = self._parse_response(resp)
        if "access_token" not in data:
            raise PyFacebookException(ErrorMessage(
                code=ErrorCode.INVALID_PARAMS,
                message=(
                    "Can not change page access token. Confirm: \n"
                    "1. Your user access token has `page_show_list` or `manage_pages` permission.\n"
                    "2. You have the target page's manage permission."
                )
            ))
        return data["access_token"]

    def exchange_insights_token(self, page_id, access_token=None):
        """
        :param page_id:
        :param access_token:
        :return:
        """
        warnings.warn(
            "This method will deprecated soon, please use exchange_page_token instead",
            PyFacebookDeprecationWaring
        )
        return self.exchange_page_token(page_id=page_id, access_token=access_token)



import unittest

from pyfacebook.ratelimit import RateLimit


class RateLimitTest(unittest.TestCase):

    def testParseHeaders(self):
        headers = {"key": "None"}
        self.assertIsNone(RateLimit.parse_headers(headers, key="key"))

    def testAppLimit(self):
        headers = {'x-app-usage': '{"call_count":91,"total_cputime":15,"total_time":12}'}
        r = RateLimit()
        r.set_limit(headers)

        self.assertEqual(r.get_limit().call_count, 91)
        self.assertEqual(r.get_limit().max_percent(), 91)
        self.assertEqual(r.get_max_percent(), 91)
        self.assertEqual(r.get_sleep_seconds(), 2)

    def testBusinessLimit(self):
        r = RateLimit()

        headers = {
            "x-business-use-case-usage": "{\"112130216863063\":[{\"type\":\"pages\",\"call_count\":1,\"total_cputime\":1,\"total_time\":1,\"estimated_time_to_regain_access\":0}]}"}

        r.set_limit(headers)

        self.assertEqual(r.get_limit(object_id="112130216863063", endpoint="pages").call_count, 1)



"""
    This is default fields for to retrieve data from Facebook graph API.

    Updated at: v5.0
"""

FB_PAGE_FIELDS = [
    "id", "about", "can_checkin", "category", "category_list", "checkins",
    "contact_address", "cover", "current_location", "description", "description_html",
    "display_subtext", "emails", "engagement", "fan_count", "founded", "general_info",
    "global_brand_page_name", "global_brand_root_id", "link", "name",
    "phone", "picture", "rating_count", "single_line_address", "start_info",
    "talking_about_count", "username", "verification_status", "website",
    "were_here_count", "whatsapp_number",
]

FB_POST_ATTACHMENTS = [
    "description", "description_tags", "media", "media_type", "target", "title",
    "type", "unshimmed_url", "url", "subattachments"
]

FB_POST_BASIC_FIELDS = [
    "id", "attachments{{{}}}".format(",".join(FB_POST_ATTACHMENTS)),
    "created_time", "full_picture", "icon", "message", "message_tags",
    "permalink_url", "picture", "shares", "status_type", "updated_time",
    # filter stream return all comments count
    # refer: https://developers.facebook.com/docs/graph-api/reference/page-post/comments/
    "comments.filter(stream).summary(true).limit(0)",
]

FB_POST_REACTIONS_FIELD = [
    "reactions.summary(true).limit(0)",
    "reactions.type(LIKE).limit(0).summary(total_count).as(like)",
    "reactions.type(LOVE).limit(0).summary(total_count).as(love)",
    "reactions.type(WOW).limit(0).summary(total_count).as(wow)",
    "reactions.type(HAHA).limit(0).summary(total_count).as(haha)",
    "reactions.type(SAD).limit(0).summary(total_count).as(sad)",
    "reactions.type(ANGRY).limit(0).summary(total_count).as(angry)",
    "reactions.type(THANKFUL).limit(0).summary(total_count).as(thankful)"
]

FB_COMMENT_BASIC_FIELDS = [
    "id", "attachment", "created_time", "like_count",
    "can_comment", "can_like", "comment_count",
    "from", "message", "permalink_url",
]

FB_PAGE_PICTURE_TYPE = [
    "small", "normal", "album", "large", "square"
]

FB_VIDEO_BASIC_FIELDS = [
    "id", "title", "description",
    "created_time", "updated_time",
    # common connections
    "likes.summary(true).limit(0)", "comments.summary(true).limit(0)"
]

FB_VIDEO_CAPTION_BASIC_FIELDS = [
    "create_time", "is_auto_generated", "is_default",
    "locale", "locale_name", "uri",
]

FB_ALBUM_BASIC_FIELDS = [
    "id", "created_time", "count",
    "name", "description",
    # common connections
    "likes.summary(true).limit(0)",
    "comments.summary(true).limit(0)",
]

FB_PHOTO_BASIC_FIELDS = [
    "id", "created_time", "name", "link",
    # common connections
    "likes.summary(true).limit(0)",
]

FB_LIVE_VIDEO_BASIC_FIELDS = [
    "id", "creation_time", "description", "title",
    "permalink_url", "live_views", "status", "secure_stream_url",
    # common connections
    "comments.summary(true).limit(0)",
]

FB_LIVE_VIDEO_INPUT_STREAM_BASIC_FIELDS = [
    "id", "dash_ingest_url", "dash_preview_url", "is_master",
    "live_encoder", "secure_stream_url", "stream_health",
    "stream_id", "stream_url",
]

# ====================================================
# Follows is Instagram Professional fields to get data
# ====================================================

INSTAGRAM_USER_FIELD = [
    "biography", "id", "ig_id",
    "followers_count", "follows_count", "media_count",
    "name", "profile_picture_url", "username", "website",
]

INSTAGRAM_MEDIA_CHILDREN_PUBLIC_FIELD = [
    "id", "media_type", "media_url", "permalink", "timestamp", "username"
]

INSTAGRAM_MEDIA_CHILDREN_OWNER_FIELD = INSTAGRAM_MEDIA_CHILDREN_PUBLIC_FIELD + [
    "ig_id", "owner", "shortcode", "thumbnail_url",
]

INSTAGRAM_MEDIA_FIELD = [
    "caption", "comments_count", "id",
    "like_count", "media_type", "media_url",
    "permalink", "timestamp", "username"
]

INSTAGRAM_MEDIA_PUBLIC_FIELD = INSTAGRAM_MEDIA_FIELD + [
    "children{{{}}}".format(",".join(INSTAGRAM_MEDIA_CHILDREN_PUBLIC_FIELD))
]

INSTAGRAM_MEDIA_OWNER_FIELD = INSTAGRAM_MEDIA_FIELD + [
    "children{{{}}}".format(",".join(INSTAGRAM_MEDIA_CHILDREN_OWNER_FIELD)),
    "ig_id", "is_comment_enabled",
    "owner", "shortcode", "thumbnail_url"
]

INSTAGRAM_STORY_FIELD = [
    "id", "ig_id", "caption", "comments_count", "like_count",
    "media_product_type", "media_type", "media_url", "owner", "permalink",
    "shortcode", "thumbnail_url", "timestamp", "username", "video_title",
]

INSTAGRAM_COMMENT_FIELD = [
    "hidden", "id", "like_count", "media",
    "text", "timestamp", "user", "username"
]

INSTAGRAM_REPLY_FIELD = INSTAGRAM_COMMENT_FIELD

INSTAGRAM_HASHTAG_FIELD = [
    "id", "name"
]

INSTAGRAM_HASHTAG_MEDIA_FIELD = [
    "caption", "children{id,media_type,media_url,permalink}",
    "comments_count", "id", "like_count", "media_type",
    "media_url", "permalink",
]

INSTAGRAM_MENTION_COMMENT_FIELD = [
    "id", "like_count", "media{{{}}}".format(",".join(INSTAGRAM_MEDIA_FIELD)),
    "text", "timestamp",
]

INSTAGRAM_CONTAINER_FIELDS = [
    "id", "status_code",
]

INSTAGRAM_PUBLISH_FIELDS = [
    "quota_usage", "config"
]

# =====================================================
# Follows is Instagram Basic display fields to get data
# =====================================================
INSTAGRAM_BASIC_USER_FIELD = [
    "account_type", "id", "media_count", "username"
]

INSTAGRAM_BASIC_MEDIA_CHILDREN_FIELD = [
    "id", "media_type", "media_url", "permalink",
    "thumbnail_url", "timestamp", "username"
]

INSTAGRAM_BASIC_MEDIA_FIELD = [
    "caption", "id", "media_type", "media_url", "permalink",
    "thumbnail_url", "timestamp", "username", "children{{{}}}".format(",".join(INSTAGRAM_BASIC_MEDIA_CHILDREN_FIELD))
]



import json
import unittest

import pyfacebook.models as models


class AccessTokenModelTest(unittest.TestCase):
    BASE_PATH = "testdata/facebook/models/access_tokens/"

    with open(BASE_PATH + 'auth_access_token.json', 'rb') as f:
        AUTH_ACCESS_TOKEN_INFO = json.loads(f.read().decode('utf-8'))
    with open(BASE_PATH + 'access_token_correct.json', 'rb') as f:
        ACCESS_TOKEN_CORRECT = json.loads(f.read().decode('utf-8'))
    with open(BASE_PATH + 'access_token_wrong.json', 'rb') as f:
        ACCESS_TOKEN_WRONG = json.loads(f.read().decode('utf-8'))

    def testAuthAccessToken(self):
        m = models.AuthAccessToken.new_from_json_dict(self.AUTH_ACCESS_TOKEN_INFO)

        self.assertEqual(m.access_token, "access_token")
        self.assertEqual(m.expires_in, 3600)

        data_dict = m.as_dict()
        self.assertEqual(data_dict["access_token"], "access_token")
        data_str = m.as_json_str()
        self.assertEqual(
            data_str,
            '{"access_token": "access_token", "token_type": "type", "expires_in": 3600, "expires_at": 1568901737}'
        )

    def testCorrectAccessToken(self):
        m = models.AccessToken.new_from_json_dict(self.ACCESS_TOKEN_CORRECT)

        self.assertEqual(m.app_id, "1234567890")
        self.assertEqual(m.metadata, None)
        self.assertEqual(len(m.scopes), 8)
        self.assertEqual(m.scopes[0], "email")
        self.assertEqual(len(m.granular_scopes), 5)
        self.assertEqual(m.granular_scopes[0].scope, "manage_pages")
        self.assertEqual(m.granular_scopes[3].target_ids[0], "123456789")

    def testWrongAccessToken(self):
        m = models.AccessToken.new_from_json_dict(self.ACCESS_TOKEN_WRONG)

        self.assertEqual(m.error.code, 190)
        self.assertEqual(m.is_valid, False)



import json

import unittest
from six import iteritems

import responses
import pyfacebook


class ApiCommentTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram/apidata/comments/"
    BASE_URL = "https://graph.facebook.com/{}/".format(pyfacebook.Api.VALID_API_VERSIONS[-1])

    with open(BASE_PATH + "comment_default_p1.json", "rb") as f:
        COMMENTS_DEFAULT_p1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comment_default_p2.json", "rb") as f:
        COMMENTS_DEFAULT_p2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comment_fields.json", "rb") as f:
        COMMENTS_FIELDS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comment_single_default.json", "rb") as f:
        COMMENT_SINGLE_DEFAULT = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comment_single_fields.json", "rb") as f:
        COMMENT_SINGLE_FIELDS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comments_multi_default.json", "rb") as f:
        COMMENTS_MULTI_DEFAULT = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comments_multi_fields.json", "rb") as f:
        COMMENTS_MULTI_FIELDS = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.api = pyfacebook.IgProApi(
            app_id="123456", app_secret="secret",
            long_term_token="token",
        )

    def testGetCommentsByMedia(self):
        media_id = "17955956875141196"
        # test get all comments
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + media_id + "/comments", json=self.COMMENTS_DEFAULT_p1)
            m.add("GET", self.BASE_URL + media_id + "/comments", json=self.COMMENTS_DEFAULT_p2)

            res = self.api.get_comments_by_media(
                media_id=media_id,
                count=None
            )
            self.assertEqual(len(res), 9)
            self.assertEqual(res[0].replies[0].id, "17850033184810160")

        # test count
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + media_id + "/comments", json=self.COMMENTS_DEFAULT_p1)
            m.add("GET", self.BASE_URL + media_id + "/comments", json=self.COMMENTS_DEFAULT_p2)

            res = self.api.get_comments_by_media(
                media_id=media_id,
                count=8,
                return_json=True
            )
            self.assertEqual(len(res), 8)
            self.assertEqual(res[6]["replies"]["data"][0]["id"], "18107567341036926")

        # test fields
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + media_id + "/comments", json=self.COMMENTS_FIELDS)

            res = self.api.get_comments_by_media(
                media_id=media_id,
                fields=["id", "like_count", "timestamp", "text"],
                count=4
            )

            self.assertEqual(len(res), 4)
            self.assertEqual(res[0].id, "17862949873623188")

    def testGetCommentInfo(self):
        comment_id = "17862949873623188"

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + comment_id, json=self.COMMENT_SINGLE_DEFAULT)
            m.add("GET", self.BASE_URL + comment_id, json=self.COMMENT_SINGLE_DEFAULT)

            res_default_with_reply = self.api.get_comment_info(
                comment_id=comment_id
            )
            self.assertEqual(res_default_with_reply.id, comment_id)
            self.assertEqual(res_default_with_reply.replies[0].id, "17850033184810160")

            res_fields = self.api.get_comment_info(
                comment_id=comment_id,
                fields=["id", "like_count", "timestamp", "text"],
                include_reply=False,
                return_json=True
            )
            self.assertEqual(res_fields["id"], comment_id)

    def testGetCommentsInfo(self):
        comment_ids = ["17862949873623188", "17844360649889631"]

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL, json=self.COMMENTS_MULTI_DEFAULT)
            m.add("GET", self.BASE_URL, json=self.COMMENTS_MULTI_FIELDS)

            res_default_with_reply = self.api.get_comments_info(
                comment_ids=comment_ids
            )
            for _id, data in iteritems(res_default_with_reply):
                self.assertIn(_id, comment_ids)
                self.assertIn(_id, data.id)

            res_fields = self.api.get_comments_info(
                comment_ids=comment_ids,
                fields=["id", "like_count", "timestamp", "text"],
                include_reply=False,
                return_json=True,
            )
            for _id, data in iteritems(res_fields):
                self.assertIn(_id, comment_ids)
                self.assertIn(_id, data["id"])



import json
import unittest

import responses
from six import iteritems

import pyfacebook


class CommentTestApi(unittest.TestCase):
    BASE_PATH = "testdata/facebook/apidata/pictures/"
    BASE_URL = "https://graph.facebook.com/v8.0/"

    with open(BASE_PATH + "single_picture.json", "rb") as f:
        SINGLE_PICTURE = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "multi_pictures.json", "rb") as f:
        MULTI_PICTURE = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.api = pyfacebook.Api(
            app_id="123456", app_secret="secret",
            long_term_token="token", version="v8.0",
        )

    def testGetPictureInfo(self):
        page_id = "2121008874780932"
        with self.assertRaises(pyfacebook.PyFacebookException):
            self.api.get_picture(page_id=page_id, pic_type="not")

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + page_id + "/picture", json=self.SINGLE_PICTURE)

            picture = self.api.get_picture(
                page_id=page_id,
            )
            self.assertEqual(picture.height, 200)

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + page_id + "/picture", json=self.SINGLE_PICTURE)

            picture = self.api.get_picture(
                page_id=page_id, pic_type="large",
                return_json=True
            )
            self.assertEqual(picture["height"], 200)

    def testGetPictures(self):
        ids = ["2121008874780932", "20531316728"]

        with self.assertRaises(pyfacebook.PyFacebookException):
            self.api.get_pictures(ids=ids, pic_type="not")

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + "picture", json=self.MULTI_PICTURE)

            picture_dict = self.api.get_pictures(
                ids=ids
            )

            for _id, data in iteritems(picture_dict):
                self.assertIn(_id, ids)
                self.assertEqual(200, data.height)

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + "picture", json=self.MULTI_PICTURE)

            picture_dict = self.api.get_pictures(
                ids=ids,
                pic_type="large",
                return_json=True
            )

            for _id, data in iteritems(picture_dict):
                self.assertIn(_id, ids)
                self.assertEqual(200, data["height"])



import json
import unittest

import responses
from six import iteritems

import pyfacebook


class ApiMediaTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram/apidata/medias/"
    BASE_URL = "https://graph.facebook.com/{}/".format(pyfacebook.Api.VALID_API_VERSIONS[-1])

    with open(BASE_PATH + "media_default_fields_p1.json", "rb") as f:
        MEDIAS_DEFAULT_p1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_default_fields_p2.json", "rb") as f:
        MEDIAS_DEFAULT_p2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_default_fields_p3.json", "rb") as f:
        MEDIAS_DEFAULT_p3 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_fields_p1.json", "rb") as f:
        MEDIAS_FIELDS_p1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_fields_p2.json", "rb") as f:
        MEDIAS_FIELDS_p2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_default.json", "rb") as f:
        MEDIA_DEFAULT = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_fields.json", "rb") as f:
        MEDIA_FIELDS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "medias_default.json", "rb") as f:
        MEDIAS_DEFAULT = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "medias_fields.json", "rb") as f:
        MEDIAS_FIELDS = json.loads(f.read().decode("utf-8"))

    with open(BASE_PATH + "tags_medias.json", "rb") as f:
        TAGS_MEDIAS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "tags_medias_p1.json", "rb") as f:
        TAGS_MEDIAS_P1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "tags_medias_p2.json", "rb") as f:
        TAGS_MEDIAS_P2 = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.instagram_business_id = "17841406338772941"
        self.api = pyfacebook.IgProApi(
            app_id="123456", app_secret="secret",
            long_term_token="token"
        )

    def testGetMedias(self):
        # test all items
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/media", json=self.MEDIAS_DEFAULT_p1)
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/media", json=self.MEDIAS_DEFAULT_p2)
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/media", json=self.MEDIAS_DEFAULT_p3)

            res = self.api.get_user_medias(
                user_id=self.instagram_business_id,
                count=None,
                limit=5,
                return_json=True
            )
            self.assertEqual(len(res), 8)

        # test count
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/media", json=self.MEDIAS_FIELDS_p1)
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/media", json=self.MEDIAS_FIELDS_p2)

            res = self.api.get_user_medias(
                user_id=self.instagram_business_id,
                fields=["id", "caption", "media_url", "like_count"],
                count=7,
                limit=5
            )
            self.assertEqual(len(res), 7)

        # test time filter
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/media", json=self.MEDIAS_DEFAULT_p1)

            res = self.api.get_user_medias(
                user_id=self.instagram_business_id,
                since="2019-10-1",
                until="2019-11-1",
                count=2,
                limit=5
            )
            self.assertEqual(len(res), 2)

    def testGetMediaInfo(self):
        media_id = "18027939643230671"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + media_id, json=self.MEDIA_DEFAULT)
            m.add("GET", self.BASE_URL + media_id, json=self.MEDIA_FIELDS)

            res_default = self.api.get_media_info(
                media_id=media_id
            )
            self.assertEqual(res_default.id, media_id)
            self.assertEqual(res_default.owner.id, self.instagram_business_id)

            res_fields = self.api.get_media_info(
                media_id=media_id,
                fields=["id", "timestamp", "media_type", "comments_count"],
                return_json=True
            )
            self.assertEqual(res_fields["id"], media_id)
            self.assertEqual(res_fields["comments_count"], 0)

    def testGetMediasInfo(self):
        media_ids = ["17861821972334188", "17864312515295083"]
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL, json=self.MEDIAS_DEFAULT)
            m.add("GET", self.BASE_URL, json=self.MEDIAS_FIELDS)

            res_default = self.api.get_medias_info(
                media_ids=media_ids
            )
            for _id, data in iteritems(res_default):
                self.assertIn(_id, media_ids)
                self.assertIn(_id, data.id)

            res_fields = self.api.get_medias_info(
                media_ids=media_ids,
                fields=["id", "timestamp", "media_type", "comments_count"],
                return_json=True
            )
            for _id, data in iteritems(res_fields):
                self.assertIn(_id, media_ids)
                self.assertIn(_id, data["id"])

    def testGetTagsMedias(self):
        # test all items
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/tags", json=self.TAGS_MEDIAS)
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/tags", json=self.TAGS_MEDIAS_P1)
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/tags", json=self.TAGS_MEDIAS_P2)

            res = self.api.get_tags_medias(
                user_id=self.instagram_business_id,
                fields=["id", "like_count", "media_type", "timestamp", "username"],
                count=4
            )

            self.assertEqual(len(res), 4)
            self.assertEqual(res[0].id, "18027939643230671")

            res_json = self.api.get_tags_medias(
                user_id=self.instagram_business_id,
                count=None,
                limit=3,
                return_json=True
            )
            self.assertEqual(len(res_json), 5)
            self.assertEqual(res_json[0]["id"], "18027939643230671")



import json
import unittest

import responses

import pyfacebook


class ApiHashtagTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram/apidata/hashtags/"
    BASE_URL = "https://graph.facebook.com/{}/".format(pyfacebook.Api.VALID_API_VERSIONS[-1])

    with open(BASE_PATH + "hashtag_search.json", "rb") as f:
        HASHTAG_SEARCH = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "recently_searched_hashtags.json", "rb") as f:
        RECENTLY_HASHTAGS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_info.json", "rb") as f:
        HASHTAG_INFO = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_top_medias_default_p1.json", "rb") as f:
        HASHTAG_TOP_MEDIAS_DEFAULT_P1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_top_medias_default_p2.json", "rb") as f:
        HASHTAG_TOP_MEDIAS_DEFAULT_P2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_top_medias_fields_p1.json", "rb") as f:
        HASHTAG_TOP_MEDIAS_FIELDS_P1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_top_medias_fields_p2.json", "rb") as f:
        HASHTAG_TOP_MEDIAS_FIELDS_P2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_recent_medias_default_p1.json", "rb") as f:
        HASHTAG_RECENT_MEDIAS_DEFAULT_P1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_recent_medias_default_p2.json", "rb") as f:
        HASHTAG_RECENT_MEDIAS_DEFAULT_P2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_recent_medias_fields_p1.json", "rb") as f:
        HASHTAG_RECENT_MEDIAS_FIELDS_P1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "hashtag_recent_medias_fields_p2.json", "rb") as f:
        HASHTAG_RECENT_MEDIAS_FIELDS_P2 = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.instagram_business_id = "123456789"
        self.api = pyfacebook.IgProApi(
            app_id="123456", app_secret="secret",
            long_term_token="token",
            instagram_business_id=self.instagram_business_id
        )

    def testSearchHashtag(self):
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + "ig_hashtag_search", json=self.HASHTAG_SEARCH)

            hashtags = self.api.search_hashtag(
                q="liukun"
            )
            self.assertEqual(len(hashtags), 1)
            self.assertEqual(hashtags[0].id, "17843421130029320")

            hashtags_json = self.api.search_hashtag(
                q="liukun",
                return_json=True
            )
            self.assertEqual(hashtags_json[0]["id"], "17843421130029320")

    def testGetHashtagInfo(self):
        hashtag_id = "17843826142012701"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + hashtag_id, json=self.HASHTAG_INFO)

            hashtag = self.api.get_hashtag_info(hashtag_id=hashtag_id)
            self.assertEqual(hashtag.id, hashtag_id)

            hashtag_json = self.api.get_hashtag_info(hashtag_id=hashtag_id, return_json=True)
            self.assertEqual(hashtag_json["name"], "love")

    def testGetHashtagTopMedias(self):
        hashtag_id = "17843826142012701"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + hashtag_id + "/top_media", json=self.HASHTAG_TOP_MEDIAS_DEFAULT_P1)
            m.add("GET", self.BASE_URL + hashtag_id + "/top_media", json=self.HASHTAG_TOP_MEDIAS_DEFAULT_P2)

            medias = self.api.get_hashtag_top_medias(
                hashtag_id=hashtag_id,
                count=None,
            )
            self.assertEqual(len(medias), 50)
            self.assertEqual(medias[0].id, "18088158883149835")

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + hashtag_id + "/top_media", json=self.HASHTAG_TOP_MEDIAS_FIELDS_P1)
            m.add("GET", self.BASE_URL + hashtag_id + "/top_media", json=self.HASHTAG_TOP_MEDIAS_FIELDS_P2)

            medias_json = self.api.get_hashtag_top_medias(
                hashtag_id=hashtag_id,
                fields=["id", "media_type", "comments_count", "like_count"],
                count=8,
                limit=5,
                return_json=True
            )
            self.assertEqual(len(medias_json), 8)
            self.assertEqual(medias_json[0]["id"], "18088158883149835")
            self.assertEqual(medias_json[0]["like_count"], 804)

    def testGetHashtagRecentMedias(self):
        hashtag_id = "17843826142012701"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + hashtag_id + "/recent_media", json=self.HASHTAG_RECENT_MEDIAS_DEFAULT_P1)
            m.add("GET", self.BASE_URL + hashtag_id + "/recent_media", json=self.HASHTAG_RECENT_MEDIAS_DEFAULT_P2)

            medias = self.api.get_hashtag_recent_medias(
                hashtag_id=hashtag_id,
                count=None,
            )
            self.assertEqual(len(medias), 50)
            self.assertEqual(medias[0].id, "17861656786698079")

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + hashtag_id + "/recent_media", json=self.HASHTAG_RECENT_MEDIAS_FIELDS_P1)
            m.add("GET", self.BASE_URL + hashtag_id + "/recent_media", json=self.HASHTAG_RECENT_MEDIAS_FIELDS_P2)

            medias_json = self.api.get_hashtag_recent_medias(
                hashtag_id=hashtag_id,
                fields=("id", "media_type", "comments_count", "like_count"),
                count=8,
                limit=5,
                return_json=True
            )
            self.assertEqual(len(medias_json), 8)
            self.assertEqual(medias_json[0]["id"], "17983329973290591")

    def testGetUserRecentlySearchedHashtags(self):
        with responses.RequestsMock() as m:
            m.add(
                "GET",
                self.BASE_URL + self.instagram_business_id + "/recently_searched_hashtags",
                json=self.RECENTLY_HASHTAGS
            )

            hashtags = self.api.get_user_recently_searched_hashtags(
                user_id=self.instagram_business_id,
                limit=3,
                access_token=self.api._access_token
            )

            self.assertEqual(len(hashtags), 3)
            self.assertEqual(hashtags[0].name, "loveyou")

            hashtags_json = self.api.get_user_recently_searched_hashtags(
                user_id=self.instagram_business_id,
                limit=3,
                access_token=self.api._access_token,
                return_json=True
            )
            self.assertEqual(hashtags_json[1]["name"], "love")



"""
    These are models for likes.
"""
from attr import attrs, attrib
from typing import Optional

from .base import BaseModel


@attrs
class LikesSummary(BaseModel):
    """
    A class representing the likes summary info.

    Refer: https://developers.facebook.com/docs/graph-api/reference/video/likes/
    """
    total_count = attrib(default=None, type=Optional[int])
    can_like = attrib(default=None, type=Optional[bool], repr=False)
    has_liked = attrib(default=None, type=Optional[bool], repr=False)



"""
    some common mixin for object
"""
from attr import attrs, attrib
from typing import Dict, Optional

from .comment import CommentSummary
from .likes import LikesSummary


@attrs
class LikesSummaryField(object):
    likes = attrib(default=None, type=Optional[Dict], repr=False)

    def __attrs_post_init__(self):
        if self.likes is not None and isinstance(self.likes, dict):
            likes_summary = self.likes.get("summary", {})
            self.likes = LikesSummary.new_from_json_dict(likes_summary)


@attrs
class CommentsSummaryField(object):
    comments = attrib(default=None, type=Optional[Dict], repr=False)

    def __attrs_post_init__(self):
        if self.comments is not None and isinstance(self.comments, dict):
            comment_summary = self.comments.get("summary", {})
            self.comments = CommentSummary.new_from_json_dict(comment_summary)


@attrs
class LikesAndCommentsSummaryField(LikesSummaryField, CommentsSummaryField):

    def __attrs_post_init__(self):
        if self.likes is not None and isinstance(self.likes, dict):
            likes_summary = self.likes.get("summary", {})
            self.likes = LikesSummary.new_from_json_dict(likes_summary)
        if self.comments is not None and isinstance(self.comments, dict):
            comment_summary = self.comments.get("summary", {})
            self.comments = CommentSummary.new_from_json_dict(comment_summary)



"""
    models for live
"""

from typing import List, Optional
from attr import attrs, attrib

from .base import BaseModel
from .mixins import CommentsSummaryField
from .._compat import str


@attrs
class LiveVideStreamHealth(BaseModel):
    """
    A class representing the live video steam heath info.

    Refer:
    """
    video_bitrate = attrib(default=None, type=Optional[float], repr=False)
    video_framerate = attrib(default=None, type=Optional[float], repr=False)
    video_gop_size = attrib(default=None, type=Optional[int], repr=False)
    video_height = attrib(default=None, type=Optional[int], repr=False)
    video_width = attrib(default=None, type=Optional[int], repr=False)
    audio_bitrate = attrib(default=None, type=Optional[float], repr=False)


@attrs
class LiveVideoInputStream(BaseModel):
    """
    A class representing the live video ingest stream info.

    Refer: https://developers.facebook.com/docs/graph-api/reference/live-video-input-stream/
    """
    id = attrib(default=None, type=Optional[str])
    dash_ingest_url = attrib(default=None, type=Optional[str], repr=False)
    dash_preview_url = attrib(default=None, type=Optional[str], repr=False)
    is_master = attrib(default=None, type=Optional[bool], repr=False)
    # live_encoder
    secure_stream_url = attrib(default=None, type=Optional[str], repr=False)
    stream_health = attrib(default=None, type=Optional[LiveVideStreamHealth], repr=False)
    stream_id = attrib(default=None, type=Optional[str], repr=False)
    stream_url = attrib(default=None, type=Optional[str], repr=False)


@attrs
class LiveVideo(BaseModel, CommentsSummaryField):
    """
    A class representing the live video info.

    Refer: https://developers.facebook.com/docs/graph-api/reference/live-video/
    """

    id = attrib(default=None, type=Optional[str])
    broadcast_start_time = attrib(default=None, type=Optional[str], repr=False)
    creation_time = attrib(default=None, type=Optional[str], repr=False)
    dash_ingest_url = attrib(default=None, type=Optional[str], repr=False)
    dash_preview_url = attrib(default=None, type=Optional[str], repr=False)
    description = attrib(default=None, type=Optional[str], repr=False)
    embed_html = attrib(default=None, type=Optional[str], repr=False)
    # _from
    ingest_streams = attrib(default=None, type=Optional[List[LiveVideoInputStream]], repr=False)
    is_manual_mode = attrib(default=None, type=Optional[bool], repr=False)
    # live_encoders
    live_views = attrib(default=None, type=Optional[int], repr=False)
    overlay_url = attrib(default=None, type=Optional[str], repr=False)
    permalink_url = attrib(default=None, type=Optional[str])
    seconds_left = attrib(default=None, type=Optional[int], repr=False)
    secure_stream_url = attrib(default=None, type=Optional[str], repr=False)
    status = attrib(default=None, type=Optional[str], repr=False)
    stream_url = attrib(default=None, type=Optional[str], repr=False)
    title = attrib(default=None, type=Optional[str], repr=False)



"""
    Instagram Professional Api impl
"""
import datetime
from typing import List, Optional, Set, Tuple, Union
from six import iteritems

from pyfacebook.error import PyFacebookException, ErrorCode, ErrorMessage
from pyfacebook.models import (
    IgProMedia, IgProUser, IgProComment, IgProReply, IgProInsight, IgProHashtag,
    IgProStory, IgProContainer, IgProPublishLimit,
)

from pyfacebook.api.base import BaseApi
from pyfacebook.utils import constant
from pyfacebook.utils.param_validation import enf_comma_separated


class IgProApi(BaseApi):
    def __init__(self,
                 app_id=None,
                 app_secret=None,
                 short_token=None,
                 long_term_token=None,
                 instagram_business_id=None,
                 version=None,
                 timeout=None,
                 sleep_on_rate_limit=False,
                 sleep_seconds_mapping=None,
                 proxies=None,
                 debug_http=False
                 ):
        BaseApi.__init__(self,
                         app_id=app_id,
                         app_secret=app_secret,
                         short_token=short_token,
                         long_term_token=long_term_token,
                         version=version,
                         timeout=timeout,
                         sleep_on_rate_limit=sleep_on_rate_limit,
                         sleep_seconds_mapping=sleep_seconds_mapping,
                         proxies=proxies,
                         debug_http=debug_http
                         )

        self.instagram_business_id = instagram_business_id

    def discovery_user(self,
                       username,  # type: str
                       fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                       return_json=False  # type: bool
                       ):
        # type: (...) -> Union[IgProUser, dict]
        """
        Retrieve other business user basic info by username.

        Note:
            Business discovery only for username and need your business id.

        :param username: The username for you want to retrieve data.
        :param fields:Comma-separated id string for data fields which you want.
        You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return instance of IgProUser.
        Or return json data. Default is false.
        :return:
        """
        if fields is None:
            fields = constant.INSTAGRAM_USER_FIELD
        param = 'business_discovery.username({username}){{{metric}}}'.format(
            username=username,
            metric=enf_comma_separated("fields", fields)
        )
        resp = self._request(
            path='{0}/{1}'.format(self.version, self.instagram_business_id),
            args={
                'fields': param
            }
        )
        data = self._parse_response(resp)
        if return_json:
            return data['business_discovery']
        else:
            return IgProUser.new_from_json_dict(data['business_discovery'])

    def discovery_user_medias(self,
                              username,  # type: str
                              fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                              since_time=None,  # type: Optional[str]
                              until_time=None,  # type: Optional[str]
                              count=10,  # type: Optional[int]
                              limit=10,  # type: int
                              return_json=False  # type: bool
                              ):
        # type: (...) -> List[Union[IgProMedia, dict]]
        """
        Retrieve other business user's public medias.

        :param username: The username for other business user.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param since_time: Lower bound of the time range to the medias publish time.
                Format is %Y-%m-%d. If not provide, will not limit by this.
        :param until_time: Upper bound of the time range to the medias publish time.
                Format is %Y-%m-%d. If not provide, will not limit by this.
        :param count: The count is you want to retrieve medias. Default is 10.
                If you want to get all data. Set it to None.
                For now This may be not more than 10K.
        :param limit: Each request retrieve posts count from api.
                For medias it should no more than 500.
        :param return_json: Set to false will return a list instance of IgProMedia.
        Or return json data. Default is false.
        """

        try:
            if since_time is not None:
                since_time = datetime.datetime.strptime(since_time, '%Y-%m-%d')
            if until_time is not None:
                until_time = datetime.datetime.strptime(until_time, '%Y-%m-%d')
        except (ValueError, TypeError):
            raise PyFacebookException(ErrorMessage(
                code=ErrorCode.INVALID_PARAMS,
                message="since_time or until_time must format as %Y-%m-%d"
            ))

        if count is not None:
            limit = min(limit, count)

        if fields is None:
            fields = constant.INSTAGRAM_MEDIA_PUBLIC_FIELD
        fields = enf_comma_separated("fields", fields)

        if (since_time is not None or until_time is not None) and "timestamp" not in fields:
            raise PyFacebookException(ErrorMessage(
                code=ErrorCode.MISSING_PARAMS,
                message="Use the since and until must give `timestamp` field"
            ))

        args = {
            'path': '{0}/{1}'.format(self.version, self.instagram_business_id),
            'username': username,
            'limit': limit,
            "metric": enf_comma_separated("fields", fields),
        }

        medias = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                args=args,
                next_cursor=next_cursor,
                business_discovery=True
            )
            data = data.get('data', [])
            # check if the media meet the request.
            for item in data:
                begin_flag, end_flag = True, True

                if "timestamp" in item:
                    timestamp = datetime.datetime.strptime(item['timestamp'][:-5], '%Y-%m-%dT%H:%M:%S')
                    if since_time is not None:
                        begin_flag = since_time < timestamp
                    if until_time is not None:
                        end_flag = until_time > timestamp

                if all([begin_flag, end_flag]):
                    if return_json:
                        medias.append(item)
                    else:
                        medias.append(IgProMedia.new_from_json_dict(item))
                if not begin_flag:
                    next_cursor = None
                    break

            if count is not None:
                if len(medias) >= count:
                    medias = medias[:count]
                    break
            if next_cursor is None:
                break
        return medias

    def paged_by_cursor(self,
                        target=None,
                        resource=None,
                        args=None,
                        next_cursor=None,
                        business_discovery=False):
        """
        Paging response data by cursor.
        If paged business response data. Parameter args need contain basic params like username and limit.
        If paged owner data just like normally request. Provide target,resource,query_args.
        Args:
            target (str, optional)
                The page id for which you want to get resource data.
            resource (str, optional)
                The resource string for data. Like media and so on.
            args (dict)
                For owner data, this is query params for request.
                For business discovery, this contain basic fields for username and so on.
            next_cursor (str, optional)
                The paging cursor str. It will return from the graph api.
            business_discovery (bool, optional)
                If use business discovery, this should be True.
        Returns:
            The origin data return from the graph api.
        """
        if business_discovery:
            query = "business_discovery.username({username}){{media{after}.limit({limit}){{{fields}}}}}"
            after = ""
            if next_cursor is not None:
                after = ".after({after})".format(after=next_cursor)
            fields = query.format(
                username=args['username'],
                limit=args['limit'],
                after=after,
                fields=args["metric"]
            )
            path = args['path']
            args = {'fields': fields}
        else:
            path = '{0}/{1}/{2}'.format(self.version, target, resource)
            if next_cursor is not None:
                args['after'] = next_cursor

        resp = self._request(
            path=path,
            args=args
        )

        next_cursor, previous_cursor = None, None
        data = self._parse_response(resp)
        # Note: business discover only support for media.
        if business_discovery:
            data = data['business_discovery']['media']
        if 'paging' in data:
            cursors = data['paging'].get('cursors', {})
            next_cursor = cursors.get('after')
            previous_cursor = cursors.get('before')
        return next_cursor, previous_cursor, data

    def get_user_info(self,
                      user_id,  # type: str
                      fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                      return_json=False  # type: bool
                      ):
        # type: (...) -> Optional[IgProUser, dict]
        """
        Retrieve ig user data by user id.
        :param user_id: The id for instagram business user which you want to get data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return instance of IgProUser.
                Or return json data. Default is false.
        """

        if fields is None:
            fields = constant.INSTAGRAM_USER_FIELD

        args = {
            'fields': enf_comma_separated("fields", fields),
        }

        resp = self._request(
            path='{0}/{1}'.format(self.version, user_id),
            args=args
        )
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProUser.new_from_json_dict(data)

    def get_user_medias(self,
                        user_id,  # type: str
                        fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                        since=None,  # type: Optional[str]
                        until=None,  # type: Optional[str]
                        count=10,  # type: Optional[int]
                        limit=10,  # type: int
                        return_json=False  # type: bool
                        ):
        # type: (...) -> List[Union[IgProMedia, dict]]
        """
        Retrieve ig user medias data by user id.
        :param user_id: The id for instagram business user which you want to get data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param since: A Unix timestamp that points to the start of a range of time-based data.
        :param until: A Unix timestamp that points to the end of a range of time-based data.
        :param count: The count for you want to get medias.
                Default is 10.
                If need get all, set this with None.
        :param limit: Each request retrieve medias count from api.
                For medias it should no more than 500.
        :param return_json: Set to false will return instance of IgProUser.
                Or return json data. Default is false.
        """

        if fields is None:
            fields = constant.INSTAGRAM_MEDIA_OWNER_FIELD

        if count is not None:
            limit = min(limit, count)

        args = {
            'fields': enf_comma_separated("fields", fields),
            'limit': limit,
            "since": since,
            "until": until,
        }

        medias = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=user_id,
                resource='media',
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])
            for item in data:
                if return_json:
                    medias.append(item)
                else:
                    medias.append(IgProMedia.new_from_json_dict(item))
            if count is not None:
                if len(medias) >= count:
                    medias = medias[:count]
                    break
            if next_cursor is None:
                break
        return medias

    def get_media_info(self,
                       media_id,  # type: str
                       fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                       return_json=False  # type: bool
                       ):
        # type: (...) -> Union[IgProMedia, dict]
        """
        Retrieve the media info by media id.
        :param media_id: The media id for which you want to get data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return instance of IgProUser.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_MEDIA_OWNER_FIELD

        args = {'fields': enf_comma_separated("fields", fields)}

        resp = self._request(
            path='{0}/{1}'.format(self.version, media_id),
            args=args
        )

        data = self._parse_response(resp)
        if return_json:
            return data
        else:
            return IgProMedia.new_from_json_dict(data)

    def get_medias_info(self,
                        media_ids,  # type: Union[str, List, Tuple, Set]
                        fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                        return_json=False  # type: bool
                        ):
        # type: (...) -> dict
        """
        Retrieve the media info by media id.
        :param media_ids: Comma-separated id string for media which you want.
                You can also pass this with an id list, tuple, set.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return a dict values are instance of IgProUser.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_MEDIA_OWNER_FIELD

        args = {
            "fields": enf_comma_separated("fields", fields),
            "ids": enf_comma_separated("media_ids", media_ids)
        }

        resp = self._request(
            path='{0}/'.format(self.version),
            args=args
        )

        data = self._parse_response(resp)
        if return_json:
            return data
        else:
            return {_id: IgProMedia.new_from_json_dict(p_data) for _id, p_data in iteritems(data)}

    def get_tags_medias(self,
                        user_id,  # type: str
                        fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                        count=10,  # type: Optional[int]
                        limit=10,  # type: int
                        return_json=False  # type: bool
                        ):
        # type: (...) -> List[Union[IgProMedia, dict]]
        """
        Retrieve a collection of IG Media objects in which an IG User has been tagged by another Instagram user.
        :param user_id: The id for instagram business user which you want to get data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param count: The count for you want to get stories.
                Default is 10.
                If need get all, set this with None.
        :param limit: Each request retrieve stories count from api.
        :param return_json: Set to false will return instance of IgProMedia.
                Or return json data. Default is false.
        :return: IgProMedia instance list or original data dict
        """
        if fields is None:
            fields = constant.INSTAGRAM_MEDIA_PUBLIC_FIELD

        if count is not None:
            limit = min(limit, count)

        args = {
            "fields": enf_comma_separated("fields", fields),
            "limit": limit
        }

        medias = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=user_id,
                resource='tags',
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])
            if return_json:
                medias += data
            else:
                medias += [IgProMedia.new_from_json_dict(item) for item in data]
            if count is not None:
                if len(medias) >= count:
                    medias = medias[:count]
                    break
            if next_cursor is None:
                break
        return medias

    def get_user_stories(self,
                         user_id,  # type: str
                         fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                         count=10,  # type: Optional[int]
                         limit=10,  # type: int
                         return_json=False  # type: bool
                         ):
        # type: (...) -> List[Union[IgProStory, dict]]
        """
        Retrieve ig user stories data by user id.
        :param user_id: The id for instagram business user which you want to get data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param count: The count for you want to get stories.
                Default is 10.
                If need get all, set this with None.
        :param limit: Each request retrieve stories count from api.
        :param return_json: Set to false will return instance of IgProStory.
                Or return json data. Default is false.
        :return: IgProStory instance list or original data list.
        """
        if fields is None:
            fields = constant.INSTAGRAM_STORY_FIELD

        if count is not None:
            limit = min(limit, count)

        args = {
            'fields': enf_comma_separated("fields", fields),
            'limit': limit
        }

        stories = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=user_id,
                resource='stories',
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])
            if return_json:
                stories += data
            else:
                stories += [IgProStory.new_from_json_dict(item) for item in data]
            if count is not None:
                if len(stories) >= count:
                    stories = stories[:count]
                    break
            if next_cursor is None:
                break
        return stories

    def get_story_info(self,
                       story_id,  # type: str
                       fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                       return_json=False  # type: bool
                       ):
        # type: (...) -> Union[IgProMedia, dict]
        """
        Retrieve the story info by story id.
        :param story_id: The story id for which you want to get data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return instance of IgProUser.
                Or return json data. Default is false.
        :return: story instance or dict info.
        """
        if fields is None:
            fields = constant.INSTAGRAM_STORY_FIELD

        args = {'fields': enf_comma_separated("fields", fields)}

        resp = self._request(
            path='{0}/{1}'.format(self.version, story_id),
            args=args
        )

        data = self._parse_response(resp)
        if return_json:
            return data
        else:
            return IgProStory.new_from_json_dict(data)

    def get_stories_info(self,
                         story_ids,  # type: Union[str, List, Tuple, Set]
                         fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                         return_json=False  # type: bool
                         ):
        # type: (...) -> dict
        """
        Retrieve the stories info by multi story id.
        :param story_ids: Comma-separated id string for story which you want.
                You can also pass this with an id list, tuple, set.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return a dict values are instance of IgProUser.
                Or return json data. Default is false.
        :return: dict of response
        """
        if fields is None:
            fields = constant.INSTAGRAM_STORY_FIELD

        args = {
            "fields": enf_comma_separated("fields", fields),
            "ids": enf_comma_separated("story_ids", story_ids)
        }

        resp = self._request(
            path='{0}/'.format(self.version),
            args=args
        )

        data = self._parse_response(resp)
        if return_json:
            return data
        else:
            return {_id: IgProStory.new_from_json_dict(p_data) for _id, p_data in iteritems(data)}

    def get_comments_by_media(self,
                              media_id,  # type: str
                              fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                              count=10,  # type: Optional[int]
                              limit=10,  # type: int
                              include_reply=True,  # type: bool
                              return_json=False  # type: bool
                              ):
        # type: (...) -> List[Union[IgProComment, dict]]
        """
        Retrieve comments data for given media id.
        :param media_id: The media id for which you want to get comment data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param count: The count for you want to get comments.
                Default is 10.
                If need get all, set this with None.
        :param limit: Each request retrieve comments count from api.
                For comments it should no more than 50.
        :param include_reply: Set to True will include the replies to the comment.
                Default is True.
        :param return_json: Set to false will return instance of IgProComment.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_COMMENT_FIELD

        if count is None:
            limit = 50  # Each query will return a maximum of 50 comments.
        else:
            limit = min(count, limit)

        fields = enf_comma_separated("fields", fields)
        if include_reply:
            fields = fields + ",replies{{{}}}".format(','.join(constant.INSTAGRAM_REPLY_FIELD))
        args = {
            'fields': fields,
            'limit': limit
        }

        comments = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=media_id,
                resource='comments',
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])

            if return_json:
                comments += data
            else:
                comments += [IgProComment.new_from_json_dict(item) for item in data]

            if count is not None:
                if len(comments) >= count:
                    comments = comments[:count]
                    break
            if next_cursor is None:
                break

        return comments

    def get_comment_info(self,
                         comment_id,  # type: str
                         fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                         include_reply=True,  # type: bool
                         return_json=False  # type: bool
                         ):
        # type: (...) -> Union[IgProComment, dict]
        """
        Retrieve comment info by the comment id.
        :param comment_id: The comment id for which you want to get comment data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param include_reply: Set to True will include the replies to the comment.
                Default is True.
        :param return_json: Set to false will return instance of IgProComment.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_COMMENT_FIELD

        fields = enf_comma_separated("fields", fields)

        if include_reply:
            fields = fields + ",replies{{{}}}".format(','.join(constant.INSTAGRAM_REPLY_FIELD))

        args = {'fields': fields}

        resp = self._request(
            path='{0}/{1}'.format(self.version, comment_id),
            args=args
        )

        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProComment.new_from_json_dict(data)

    def get_comments_info(self,
                          comment_ids,  # type: Union[str, List, Tuple, Set]
                          fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                          include_reply=True,  # type: bool
                          return_json=False  # type: bool
                          ):
        # type: (...) -> dict
        """
        Retrieve comment info by the comment id.
        :param comment_ids: Comma-separated id string for comment which you want.
                You can also pass this with an id list, tuple, set.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param include_reply: Set to True will include the replies to the comment.
                Default is True.
        :param return_json: Set to false will return a dict values are instance of IgProComment.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_COMMENT_FIELD

        fields = enf_comma_separated("fields", fields)

        if include_reply:
            fields = fields + ",replies{{{}}}".format(','.join(constant.INSTAGRAM_REPLY_FIELD))

        args = {
            'fields': fields,
            "ids": enf_comma_separated("comment_ids", comment_ids)
        }

        resp = self._request(
            path='{0}/'.format(self.version),
            args=args
        )

        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return {_id: IgProComment.new_from_json_dict(p_data) for _id, p_data in iteritems(data)}

    def get_replies_by_comment(self,
                               comment_id,  # type: str
                               fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                               count=10,  # type: Optional[int]
                               limit=10,  # type: int
                               return_json=False  # type: bool
                               ):
        # type: (...) -> List[Union[IgProReply, dict]]
        """
        Retrieve replies for the given comment.
        :param comment_id: The comment id for which you want to get replies data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param count: The count for you want to get replies.
                Default is 10.
                If need get all, set this with None.
        :param limit: Each request retrieve replies count from api.
                For replies it should no more than 100.
        :param return_json: Set to false will return a list of instance of IgProReply.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_REPLY_FIELD

        if count is not None:
            limit = min(count, limit)

        args = {
            'fields': enf_comma_separated("fields", fields),
            'limit': limit
        }

        replies = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=comment_id,
                resource='replies',
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])

            if return_json:
                replies += data
            else:
                replies += [IgProReply.new_from_json_dict(item) for item in data]
            if next_cursor is None:
                break
            if count is not None:
                if len(replies) >= count:
                    replies = replies[:count]
                    break
        return replies

    def get_reply_info(self,
                       reply_id,  # type: str
                       fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                       return_json=False  # type: bool
                       ):
        # type: (...) -> Union[IgProReply, dict]
        """
        Retrieve reply info by reply id.
        :param reply_id: The reply id for which you want to get reply data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return instance of IgProComment.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_REPLY_FIELD

        args = {
            'fields': enf_comma_separated("fields", fields),
        }

        resp = self._request(
            path='{0}/{1}'.format(self.version, reply_id),
            args=args
        )
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProReply.new_from_json_dict(data)

    def get_replies_info(self,
                         reply_ids,  # type: Union[str, List, Tuple, Set]
                         fields=None,  # type: Optional[Union[str, List, Tuple, Set]]
                         return_json=False  # type: bool
                         ):
        # type: (...) -> dict
        """
        Retrieve reply info by reply id.
        :param reply_ids: Comma-separated id string for reply which you want.
                You can also pass this with an id list, tuple, set.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param return_json: Set to false will return a dict of values are instance of IgProComment.
                Or return json data. Default is false.
        """
        if fields is None:
            fields = constant.INSTAGRAM_REPLY_FIELD

        args = {
            'fields': enf_comma_separated("fields", fields),
            "ids": enf_comma_separated("reply_ids", reply_ids)
        }

        resp = self._request(
            path='{0}/'.format(self.version),
            args=args
        )
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return {_id: IgProReply.new_from_json_dict(p_data) for _id, p_data in iteritems(data)}

    def get_user_insights(self,
                          user_id,  # type: str
                          period,  # type: str
                          metrics,  # type: Union[str, List, Tuple, Set]
                          since=None,  # type: Optional[int],
                          until=None,  # type: Optional[int],
                          access_token=None,  # type: str
                          return_json=False,  # type: bool
                          ):
        # type: (...) -> List[Union[IgProInsight, dict]]
        """
        Retrieve instagram business account user insights data.
        :param user_id: The id for instagram business user which you want to get data.
        :param period: The period to aggregation data.
                Accepted parameters:
                    - lifetime
                    - day
                    - days_28
        :param metrics: Comma-separated id string for metrics that needs to be fetched..
                You can also pass this with an id list, tuple, set.
                Note:
                    some metrics incompatible with the period.
                    see more: https://developers.facebook.com/docs/instagram-api/reference/user/insights#metrics-periods
        :param since: Lower bound of the time range to fetch data. Need Unix timestamps.
        :param until: Upper bound of the time range to fetch data. Need Unix timestamps.
                The time range not more than 30 days (2592000 s).
        :param access_token: Target user access token. If not will use default access token.
        :param return_json: Set to false will return a list of instance of IgProInsight.
                Or return json data. Default is false.
        """

        args = {
            "metric": enf_comma_separated("metrics", metrics),
            "period": period
        }

        if access_token is not None:
            args["access_token"] = access_token

        if since is not None:
            args["since"] = since
        if until is not None:
            args["until"] = until

        resp = self._request(
            path="{0}/{1}/insights".format(self.version, user_id),
            args=args
        )

        data = self._parse_response(resp)
        if return_json:
            return data["data"]
        else:
            return [IgProInsight.new_from_json_dict(item) for item in data["data"]]

    def get_media_insights(self,
                           media_id,  # type: str
                           metrics,  # type: Union[str, List, Tuple, Set]
                           access_token=None,  # type: str
                           return_json=False,  # type: bool
                           ):
        # type: (...) -> List[Union[IgProInsight, dict]]
        """
        Retrieve given media insights data.
        :param media_id: The media id for which you want to get data.
        :param metrics: Comma-separated id string for metrics that needs to be fetched..
                You can also pass this with an id list, tuple, set.
                Note:
                    Different media type has different metric.
                    see more: https://developers.facebook.com/docs/instagram-api/reference/media/insights#insights-2
        :param access_token: Target user access token. If not will use default access token.
        :param return_json: Set to false will return a list of instance of IgProInsight.
                Or return json data. Default is false.
        """

        args = {
            "metric": enf_comma_separated("metrics", metrics)
        }

        if access_token is not None:
            args["access_token"] = access_token

        resp = self._request(
            path="{0}/{1}/insights".format(self.version, media_id),
            args=args
        )

        data = self._parse_response(resp)
        if return_json:
            return data["data"]
        else:
            return [IgProInsight.new_from_json_dict(item) for item in data["data"]]

    def search_hashtag(self,
                       q,  # type: str
                       return_json=False,  # type: bool
                       ):
        # type: (...) ->  List[Union[IgProHashtag, dict]]
        """
        Retrieve IG Hashtag IDs.

        Note:
            You can query a maximum of 30 unique hashtags within a 7 day period.

        :param q: The hashtag name to query.
        :param return_json: Set to false will return a list of instance of IgProHashtag.
                Or return json data. Default is false.
        :return: hashtag data list
        """

        args = {
            "user_id": self.instagram_business_id,
            "q": q
        }

        resp = self._request(
            path="{0}/ig_hashtag_search".format(self.version),
            args=args
        )

        data = self._parse_response(resp)

        if return_json:
            return data["data"]
        else:
            return [IgProHashtag.new_from_json_dict(item) for item in data["data"]]

    def get_hashtag_info(self,
                         hashtag_id,  # type: str
                         return_json=False,  # type: bool
                         ):
        # type: (...) -> Optional[IgProHashtag, dict]
        """
        Retrieve hashtag info by hashtag id.

        :param hashtag_id: The id for target hashtag.
        :param return_json: Set to false will return an instance of IgProHashtag.
                Or return json data. Default is false.
        :return: Hashtag info.
        """

        args = {
            "fields": "id,name"
        }

        resp = self._request(
            path="{0}/{1}".format(self.version, hashtag_id),
            args=args
        )

        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProHashtag.new_from_json_dict(data)

    def get_hashtag_top_medias(self,
                               hashtag_id,  # type: str
                               fields=None,  # type: Union[str, List, Tuple, Set]
                               count=25,  # type: Optional[int]
                               limit=25,  # type: int
                               return_json=False,  # type: bool
                               ):
        # type: (...) -> List[Union[IgProMedia, dict]]
        """
        Retrieve the most popular photo and video IG Media objects that have been tagged with the hashtag.

        :param hashtag_id: The id for hashtag which you want to retrieve data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param count: The count for you want to get medias.
                Default is 25.
                If need get all, set this with None.
        :param limit: Each request retrieve comments count from api.
                For comments it should no more than 50.
        :param return_json: Set to false will return a list of instance of IgProMedia.
                Or return json data. Default is false.
        :return: media data list.
        """
        if fields is None:
            fields = constant.INSTAGRAM_HASHTAG_MEDIA_FIELD

        if count is None:
            limit = 50  # Each query will return a maximum of 50 medias.
        else:
            limit = min(count, limit)

        args = {
            "user_id": self.instagram_business_id,
            "fields": enf_comma_separated(field="fields", value=fields),
            "limit": limit,
        }

        medias = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=hashtag_id,
                resource='top_media',
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])

            if return_json:
                medias += data
            else:
                medias += [IgProMedia.new_from_json_dict(item) for item in data]
            if count is not None:
                if len(medias) >= count:
                    medias = medias[:count]
                    break
            if next_cursor is None:
                break
        return medias

    def get_hashtag_recent_medias(self,
                                  hashtag_id,  # type: str
                                  fields=None,  # type: Union[str, List, Tuple, Set]
                                  count=25,  # type: Optional[int]
                                  limit=25,  # type: int
                                  return_json=False,  # type: bool
                                  ):
        # type: (...) -> List[Union[IgProMedia, dict]]
        """
        Retrieve a list of the most recently published photo and video IG Media objects
        published with a specific hashtag.

        :param hashtag_id: The id for hashtag which you want to retrieve data.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
        :param count: The count for you want to get medias.
                Default is 25.
                If need get all, set this with None.
        :param limit: Each request retrieve comments count from api.
                For medias it should no more than 50.
        :param return_json: Set to false will return a list of instance of IgProMedia.
                Or return json data. Default is false.
        :return: media data list.
        """
        if fields is None:
            fields = constant.INSTAGRAM_HASHTAG_MEDIA_FIELD

        if count is None:
            limit = 50  # Each query will return a maximum of 50 medias.
        else:
            limit = min(count, limit)

        args = {
            "user_id": self.instagram_business_id,
            "fields": enf_comma_separated(field="fields", value=fields),
            "limit": limit,
        }

        medias = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=hashtag_id,
                resource='recent_media',
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])

            if return_json:
                medias += data
            else:
                medias += [IgProMedia.new_from_json_dict(item) for item in data]
            if count is not None:
                if len(medias) >= count:
                    medias = medias[:count]
                    break
            if next_cursor is None:
                break
        return medias

    def get_user_recently_searched_hashtags(self,
                                            user_id,  # type: str
                                            limit=25,  # type: int
                                            access_token=None,  # type: str
                                            return_json=False,  # type: bool
                                            ):
        # type: (...) -> List[Union[IgProHashtag, dict]]
        """
        Retrieve the IG Hashtags that an IG User has searched for within the last 7 days.

        :param user_id: Instagram business account id.
        :param limit: Each request retrieve hashtags count from api.
                For this method. limit can't more than 30.
        :param access_token: Target user access token. If not will use default access token.
        :param return_json: Set to false will return a list of instance of IgProHashtag.
                Or return json data. Default is false.
        :return: hashtag data list
        """
        args = {
            "fields": "id,name",
            "limit": limit,
        }

        if access_token is not None:
            args["access_token"] = access_token

        resp = self._request(
            path="{0}/{1}/recently_searched_hashtags".format(self.version, user_id),
            args=args
        )

        data = self._parse_response(resp)

        if return_json:
            return data["data"]
        else:
            return [IgProHashtag.new_from_json_dict(item) for item in data["data"]]

    def get_tagged_user_medias(self,
                               user_id,  # type: str
                               fields=None,  # type: Union[str, List, Tuple, Set]
                               count=50,  # type: Optional[int]
                               limit=50,  # type: int
                               access_token=None,  # type: str
                               return_json=False,  # type: bool
                               ):
        # type: (...) -> List[Union[IgProMedia, dict]]
        """
        You can use this to retrieve medias which an ig user has been tagged by another ig user.

        Note:
            The private ig media will not be returned.

        :param user_id: Target user id which result medias tagged.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
                Default is all public fields.
        :param count: The you want to get medias.
                Default is 50.
                If you want to get all medias. set with None.
        :param limit: Each request retrieve comments count from api.
                Not have a exact value. And default is 50.
        :param access_token: Target user access token. If not will use default access token.
        :param return_json: Set to false will return a list of instance of IgProMedia.
                Or return json data. Default is false.
        :return: medias list
        """
        if fields is None:
            fields = constant.INSTAGRAM_MEDIA_PUBLIC_FIELD + [
                "comments{{{}}}".format(",".join(constant.INSTAGRAM_COMMENT_FIELD))
            ]

        if count is not None:
            limit = min(count, limit)

        args = {
            "fields": enf_comma_separated(field="fields", value=fields),
            "limit": limit,
        }

        if access_token is not None:
            args["access_token"] = access_token

        medias = []
        next_cursor = None

        while True:
            next_cursor, previous_cursor, data = self.paged_by_cursor(
                target=user_id,
                resource="tags",
                args=args,
                next_cursor=next_cursor
            )
            data = data.get('data', [])

            if return_json:
                medias += data
            else:
                medias += [IgProMedia.new_from_json_dict(item) for item in data]
            if count is not None:
                if len(medias) >= count:
                    medias = medias[:count]
            if next_cursor is None:
                break
        return medias

    def get_mentioned_comment_info(self,
                                   user_id,  # type: str
                                   comment_id,  # type: str
                                   fields=None,  # type: Union[str, List, Tuple, Set]
                                   access_token=None,  # type: str
                                   return_json=False,  # type: bool
                                   ):
        # type: (...) -> Union[IgProComment, dict]
        """
        You can use this to retrieve comment data which an ig user has been @mentioned by another ig user.

        :param user_id: Target user id which comment mentioned for.
        :param comment_id: The comment id which the ig user has been @mentioned.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
                Default is all public fields. (id,like_count,text,timestamp,media)
        :param access_token: Target user access token. If not will use default access token.
        :param return_json: Set to false will return a list of instance of IgProComment.
                Or return json data. Default is false.
        :return: comment data
        """

        if fields is None:
            fields = constant.INSTAGRAM_MENTION_COMMENT_FIELD

        args = {
            "fields": "mentioned_comment.comment_id({comment_id}){{{fields}}}".format(
                comment_id=comment_id,
                fields=enf_comma_separated(field="fields", value=fields)
            ),
        }

        if access_token is not None:
            args["access_token"] = access_token

        resp = self._request(
            path="{0}/{1}".format(self.version, user_id),
            args=args
        )
        data = self._parse_response(resp)

        if return_json:
            return data["mentioned_comment"]
        else:
            return IgProComment.new_from_json_dict(data["mentioned_comment"])

    def get_mentioned_media_info(self,
                                 user_id,  # type: str
                                 media_id,  # type: str
                                 fields=None,  # type: Union[str, List, Tuple, Set]
                                 access_token=None,  # type: str
                                 return_json=False,  # type: bool
                                 ):
        # type: (...) -> Union[IgProMedia, dict]
        """
        You can use this to retrieve media info which an ig user has been @mentioned in a caption by another ig user.

        :param user_id: Target user id which media mentioned for.
        :param media_id: The media id which the ig user has been @mentioned.
        :param fields: Comma-separated id string for data fields which you want.
                You can also pass this with an id list, tuple, set.
                Default is all public fields. fields as follows:
                (caption,comments,comments_count,like_count,media_type,media_url,owner,timestamp,username)
        :param access_token: Target user access token. If not will use default access token.
        :param return_json: Set to false will return a list of instance of IgProMedia.
                Or return json data. Default is false.
        :return: media data
        """

        if fields is None:
            fields = constant.INSTAGRAM_MEDIA_PUBLIC_FIELD + [
                "comments{{{}}}".format(",".join(constant.INSTAGRAM_COMMENT_FIELD))
            ]

        args = {
            "fields": "mentioned_media.media_id({media_id}){{{fields}}}".format(
                media_id=media_id,
                fields=enf_comma_separated(field="fields", value=fields)
            )
        }

        if access_token is not None:
            args["access_token"] = access_token

        resp = self._request(
            path="{0}/{1}".format(self.version, user_id),
            args=args
        )
        data = self._parse_response(resp)

        if return_json:
            return data["mentioned_media"]
        else:
            return IgProMedia.new_from_json_dict(data["mentioned_media"])

    def get_container_info(self,
                           container_id,  # type: str,
                           fields=None,  # type: Union[str, List, Tuple, Set]
                           return_json=False  # type: bool
                           ):
        # type: (...) -> Union[IgProContainer, dict]
        """
        :param container_id:
        :param fields: A comma-separated list of fields and edges you want returned.
                If omitted, default fields will be returned.
        :param return_json: Set to false will return instance of IgProContainer.
                Or return json data. Default is false.
        :return: container info
        """

        if fields is None:
            fields = constant.INSTAGRAM_CONTAINER_FIELDS

        args = {
            "fields": enf_comma_separated(field="fields", value=fields),
        }

        resp = self._request(
            path="{0}/{1}".format(self.version, container_id),
            args=args,
        )
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProContainer.new_from_json_dict(data)

    def get_publish_limit(self,
                          fields=None,  # type: Union[str, List, Tuple, Set]
                          since=None,  # type: Optional[str]
                          return_json=False  # type: bool
                          ):
        # type: (...) -> Union[IgProPublishLimit, dict]
        """
        :param fields: A comma-separated list of fields you want returned.
                If omitted, the quota_usage field will be returned by default.
        :param since: A Unix timestamp no older than 24 hours.
        :param return_json: Set to false will return instance of IgProPublishLimit.
                Or return json data. Default is false.
        :return: publish limit info
        """

        if fields is None:
            fields = constant.INSTAGRAM_PUBLISH_FIELDS

        args = {
            "fields": enf_comma_separated(field="fields", value=fields),
            "since": since,
        }

        resp = self._request(
            path="{0}/{1}/content_publishing_limit".format(self.version, self.instagram_business_id),
            args=args
        )

        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProPublishLimit.new_from_json_dict(data["data"][0])

    def create_photo(self,
                     image_url,  # type: str
                     caption=None,  # type: Optional[str]
                     location_id=None,  # type: Optional[str]
                     user_tags=None,  # type: Optional[str]
                     return_json=False  # type: bool
                     ):
        # type: (...) -> Union[IgProContainer, dict]
        """
        :param image_url: The path to the photo. We will cURL your photo using the passed in URL
                so it must be on a public server.
        :param caption: Caption for the photo. You can include hashtags and @tags.
                Maximum 2200 characters, 30 hashtags, and 20 @ tags.
        :param location_id: The ID of a Page associated with a location that you want to tag the photo with.
        :param user_tags: An json array of public usernames and x/y coordinates for any public Instagram users who
                you want to tag in the photo.
                The array must be formatted in JSON and contain a username, x, and y, property.
        :param return_json: return_json: Set to false will return instance of IgProContainer.
                Or return json data. Default is false.
        :return: Container info.
        """

        args = {
            "image_url": image_url,
            "caption": caption,
            "location_id": location_id,
            "user_tags": user_tags,
        }

        resp = self._request(
            method="POST",
            path="{0}/media".format(self.instagram_business_id),
            args=args,
        )
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProContainer.new_from_json_dict(data)

    def create_video(self,
                     video_url,  # type: str
                     caption=None,  # type: Optional[str]
                     location_id=None,  # type: Optional[str]
                     thumb_offset=0,  # type: Optional[int]
                     return_json=False  # type: bool
                     ):
        # type: (...) -> Union[IgProContainer, dict]
        """
        :param video_url: Path to the video. We will cURL your video using the passed in URL
                so it must be on a public server.
        :param caption: Caption for the video. You can include hashtags and @tags.
                Maximum 2200 characters, 30 hashtags, and 20 @ tags.
        :param location_id: The ID of a Page associated with a location that you want to tag the video with.
        :param thumb_offset: Location, in milliseconds, of the video frame to be used as the video's cover
                thumbnail image. Default value is 0, which is the first frame of the video.
        :param return_json: return_json: Set to false will return instance of IgProContainer.
                Or return json data. Default is false.
        :return: Container info.
        """
        args = {
            "video_url": video_url,
            "caption": caption,
            "location_id": location_id,
            "thumb_offset": thumb_offset,
            "media_type": "VIDEO",
        }

        resp = self._request(
            method="POST",
            path="{0}/media".format(self.instagram_business_id),
            args=args,
        )
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return IgProContainer.new_from_json_dict(data)

    def publish_container(self, creation_id):
        # type: (str) -> Union[dict]
        """
        :param creation_id: Container's id for media which you want to publish.
        :return: data for media ids
        """
        resp = self._request(
            method="POST",
            path="{0}/media_publish".format(self.instagram_business_id),
            args={"creation_id": creation_id},
        )

        data = self._parse_response(resp)
        return data



import json
import unittest

import responses

import pyfacebook


class ApiMediaTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram_basic/apidata/media/"

    with open(BASE_PATH + "user_medias_p1.json", "rb") as f:
        USER_MEDIAS_P1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "user_medias_p2.json", "rb") as f:
        USER_MEDIAS_P2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_info.json", "rb") as f:
        MEDIA_INFO = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_children.json", "rb") as f:
        MEDIA_CHILDREN = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.api = pyfacebook.IgBasicApi(long_term_token="token")

    def testGetUserMedia(self):
        user_id = "17841406338772941"

        # test count
        with responses.RequestsMock() as m:
            m.add("GET", self.api.GRAPH_URL + "me/media", json=self.USER_MEDIAS_P1)

            medias = self.api.get_user_medias(
                count=4,
                limit=6
            )
            self.assertEqual(len(medias), 4)

        # test all medias
        with responses.RequestsMock() as m:
            m.add("GET", self.api.GRAPH_URL + user_id + "/media", json=self.USER_MEDIAS_P1)
            m.add("GET", self.api.GRAPH_URL + user_id + "/media", json=self.USER_MEDIAS_P2)

            medias_json = self.api.get_user_medias(
                user_id=user_id,
                count=None,
                limit=6,
                return_json=True
            )
            self.assertEqual(len(medias_json), 10)

    def testGetMediaInfo(self):
        media_id = "18027939643230671"

        with responses.RequestsMock() as m:
            m.add("GET", self.api.GRAPH_URL + media_id, json=self.MEDIA_INFO)

            media = self.api.get_media_info(
                media_id=media_id,
            )
            self.assertEqual(media.id, media_id)
            self.assertEqual(len(media.children), 2)

            media_json = self.api.get_media_info(
                media_id=media_id,
                return_json=True
            )
            self.assertEqual(media_json["id"], media_id)
            self.assertEqual(media_json["media_type"], "CAROUSEL_ALBUM")

    def testGetMediaChildren(self):
        media_id = "18027939643230671"

        with responses.RequestsMock() as m:
            m.add("GET", self.api.GRAPH_URL + media_id + "/children", json=self.MEDIA_CHILDREN)

            children = self.api.get_media_children(
                media_id=media_id,
            )
            self.assertEqual(len(children), 2)
            self.assertEqual(children[0].id, "17895013408404508")

            children_json_list = self.api.get_media_children(
                media_id=media_id,
                return_json=True
            )
            self.assertEqual(len(children_json_list), 2)
            self.assertEqual(children_json_list[0]["id"], "17895013408404508")



import json
import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

import responses
from requests import HTTPError

import pyfacebook
from pyfacebook.api.base import BaseApi


class BaseApiTest(unittest.TestCase):
    BASE_PATH = "testdata/base/"
    BASE_URL = "https://graph.facebook.com/{}/".format(pyfacebook.Api.VALID_API_VERSIONS[-1])
    ACCESS_TOKEN_URL = BASE_URL + "oauth/access_token"

    with open(BASE_PATH + "app_token.json", "rb") as f:
        APP_ACCESS_TOKEN = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "long_term_token.json", "rb") as f:
        LONG_TERM_TOKEN = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "token_info.json", "rb") as f:
        TOKEN_INFO = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "insights_token.json", "rb") as f:
        INSIGHTS_TOKEN = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "insights_token_none.json", "rb") as f:
        INSIGHTS_TOKEN_NONE = json.loads(f.read().decode("utf-8"))

    def testApiVersion(self):
        with self.assertRaises(pyfacebook.PyFacebookException):
            BaseApi(version="1.0")

        with self.assertRaises(pyfacebook.PyFacebookException):
            BaseApi(version="v1.0")

        with self.assertRaises(pyfacebook.PyFacebookException):
            BaseApi(version="version")

        api = BaseApi(long_term_token="token", version=None)
        self.assertEqual(api.version, api.VALID_API_VERSIONS[-1])

        api = BaseApi(long_term_token="token", version="3.3")
        self.assertEqual(api.version, "v3.3")

        api = BaseApi(long_term_token="token", version="v10.0")
        self.assertEqual(api.version, "v10.0")

    @patch('time.sleep', return_value=None)
    def testApiToken(self, patched_time_sleep):
        with self.assertRaises(pyfacebook.PyFacebookException):
            BaseApi()

        api = BaseApi(long_term_token="token", debug_http=True)
        self.assertEqual(api._access_token, "token")

        with responses.RequestsMock() as m:
            m.add("GET", self.ACCESS_TOKEN_URL, json=self.APP_ACCESS_TOKEN)
            api = BaseApi(app_id="123456789", app_secret="secret", application_only_auth=True, sleep_on_rate_limit=True)
            self.assertEqual(api._access_token, "123456789|fvYq7ORmqKa2IDCijArPOYKB0")

        with responses.RequestsMock() as m:
            m.add("GET", self.ACCESS_TOKEN_URL, json=self.LONG_TERM_TOKEN)
            api = BaseApi(app_id="123456789", app_secret="secret", short_token="short-lived token")
            self.assertEqual(api._access_token, "token")

    def testBuildSleepResource(self):
        d = {10: 3, 20: 5, 50: 20}
        api = BaseApi(long_term_token="token", sleep_on_rate_limit=True, sleep_seconds_mapping=d)
        self.assertEqual(len(api.sleep_seconds_mapping), 3)
        self.assertEqual(api.sleep_seconds_mapping[0].percent, 10)
        self.assertEqual(api.rate_limit.get_sleep_seconds(sleep_data=api.sleep_seconds_mapping), 3)

        # test max sleep
        headers = {'x-app-usage': '{"call_count":120,"total_cputime":15,"total_time":12}'}
        api.rate_limit.set_limit(headers)
        self.assertEqual(api.rate_limit.get_sleep_seconds(sleep_data=api.sleep_seconds_mapping), 60 * 10)

    def testRequest(self):
        api = BaseApi(app_id="12345678", app_secret="secret", long_term_token="token")
        with responses.RequestsMock() as m:
            m.add("GET", "https://graph.facebook.com/", json={"message": "message"})

            r = api._request(
                path="", args=None
            )
            r.json()["message"] = "message"

        with responses.RequestsMock() as m:
            m.add("POST", "https://graph.facebook.com/", json={"message": "message"})

            r = api._request(
                path="", method="POST", args=None, post_args={"auth": 1},
            )
            r.json()["message"] = "message"

        with responses.RequestsMock() as m:
            m.add("GET", "https://graph.facebook.com/", body=HTTPError("..."))

            with self.assertRaises(pyfacebook.PyFacebookException):
                api._request(
                    path="",
                )

    def testCheckError(self):
        data = {"error": {"code": 100}}
        api = BaseApi(long_term_token="token")
        with self.assertRaises(pyfacebook.PyFacebookException):
            api._check_graph_error(data)

    def testGetLongToken(self):
        api = BaseApi(app_id="id", app_secret="secret", long_term_token="token")

        with responses.RequestsMock() as m:
            m.add("GET", self.ACCESS_TOKEN_URL, json=self.LONG_TERM_TOKEN)

            token = api.get_long_token(short_token="short-lived token", return_json=True)

            self.assertEqual(token["access_token"], "token")

    def testGetAppToken(self):
        api = BaseApi(app_id="id", app_secret="secret", long_term_token="token")
        with responses.RequestsMock() as m:
            m.add("GET", self.ACCESS_TOKEN_URL, json=self.APP_ACCESS_TOKEN)

            token = api.get_app_token(return_json=True)

            self.assertEqual(token["access_token"], "123456789|fvYq7ORmqKa2IDCijArPOYKB0")

    def testGetTokenInfo(self):
        debug_token_utl = "https://graph.facebook.com/{}/debug_token".format(BaseApi.VALID_API_VERSIONS[-1])
        with responses.RequestsMock() as m:
            m.add("GET", debug_token_utl, json=self.TOKEN_INFO)

            api = BaseApi(long_term_token="long-live token")
            info = api.get_token_info(return_json=True)
            self.assertEqual(info["type"], "USER")

            api = BaseApi(app_id="appId", app_secret="Secret", long_term_token="long-live token")
            info = api.get_token_info(input_token="NeedToken")
            self.assertEqual(info.is_valid, True)

    def testGetAuthorizationUrl(self):
        with self.assertRaises(pyfacebook.PyFacebookException):
            api = BaseApi(long_term_token="token")
            api.get_authorization_url()

        api = BaseApi(app_id="appId", app_secret="appSecret", long_term_token="token")

        url, state = api.get_authorization_url()
        self.assertEqual(state, "PyFacebook")
        self.assertTrue(url)

    def testExchangeAccessToken(self):
        api = BaseApi(app_id="appId", app_secret="appSecret", long_term_token="token")
        _, state = api.get_authorization_url()

        response = "https://localhost/?code=code&state=PyFacebook#_=_"
        with responses.RequestsMock() as m:
            m.add("POST", api.exchange_access_token_url, json=self.LONG_TERM_TOKEN)

            r1 = api.exchange_access_token(response=response, return_json=True)
            self.assertEqual(r1["access_token"], "token")

            r2 = api.exchange_access_token(response=response)
            self.assertEqual(r2.access_token, "token")
            self.assertEqual(r2.token_type, "bearer")

    def testExchangeInsightsToken(self):
        page_id = "123456"
        api = pyfacebook.Api(long_term_token="token")
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + page_id, json=self.INSIGHTS_TOKEN_NONE)
            with self.assertRaises(pyfacebook.PyFacebookException):
                api.exchange_insights_token(page_id=page_id)

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + page_id, json=self.INSIGHTS_TOKEN)

            token = api.exchange_page_token(page_id=page_id)
            self.assertEqual(token, "token")

            token = api.exchange_page_token(page_id=page_id, access_token="token")
            self.assertEqual(token, "token")

            token = api.exchange_insights_token(page_id=page_id)
            self.assertEqual(token, "token")



import json
import unittest

import pyfacebook.models as models


class CommentModelTest(unittest.TestCase):
    BASE_PATH = "testdata/facebook/models/comments/"

    with open(BASE_PATH + 'comment.json', 'rb') as f:
        COMMENT_INFO = json.loads(f.read().decode('utf-8'))

    def testComment(self):
        m = models.Comment.new_from_json_dict(self.COMMENT_INFO)

        self.assertEqual(m.id, "2498598377021978_2498617433686739")
        self.assertEqual(m.comment_count, 0)
        self.assertEqual(m.attachment.media.image.height, 474)



"""
    These are models for photo entity.

    Refer: https://developers.facebook.com/docs/graph-api/reference/photo
"""

from attr import attrs, attrib
from typing import Dict, List, Optional

from .base import BaseModel
from .common import Place, EntityAtTextRange
from .mixins import LikesSummaryField
from .._compat import str


@attrs
class PlatformImageSource(BaseModel):
    height = attrib(default=None, type=Optional[int], repr=False)
    source = attrib(default=None, type=Optional[str], repr=False)
    width = attrib(default=None, type=Optional[int], repr=False)


@attrs
class Photo(BaseModel, LikesSummaryField):
    """
    A class representing the photo info.
    """

    id = attrib(default=None, type=Optional[str])
    album = attrib(default=None, type=Optional[Dict], repr=False)
    alt_text = attrib(default=None, type=Optional[str], repr=False)
    alt_text_custom = attrib(default=None, type=Optional[str], repr=False)
    backdated_time = attrib(default=None, type=Optional[str], repr=False)
    backdated_time_granularity = attrib(default=None, type=Optional[str], repr=False)
    can_backdate = attrib(default=None, type=Optional[bool], repr=False)
    can_delete = attrib(default=None, type=Optional[bool], repr=False)
    can_tag = attrib(default=None, type=Optional[bool], repr=False)
    created_time = attrib(default=None, type=Optional[str])
    # event = None
    # TODO now cattr not allow rename
    # from
    height = attrib(default=None, type=Optional[int], repr=False)
    icon = attrib(default=None, type=Optional[str], repr=False)
    images = attrib(default=None, type=Optional[List[PlatformImageSource]], repr=False)
    link = attrib(default=None, type=Optional[str], repr=False)
    name = attrib(default=None, type=Optional[str])
    name_tags = attrib(default=None, type=Optional[List[EntityAtTextRange]], repr=False)
    page_story_id = attrib(default=None, type=Optional[str], repr=False)
    place = attrib(default=None, type=Optional[Place], repr=False)
    updated_time = attrib(default=None, type=Optional[str], repr=False)
    webp_images = attrib(default=None, type=Optional[List[PlatformImageSource]], repr=False)
    width = attrib(default=None, type=Optional[int], repr=False)

    def __attrs_post_init__(self):
        super(Photo, self).__attrs_post_init__()
        if self.album is not None:
            from .album import Album
            self.album = Album.new_from_json_dict(self.album)



"""
    models for instagram basic api.
"""
from attr import attrs, attrib
from typing import Dict, Optional

from .base import BaseModel
from .._compat import str


@attrs
class IgBasicUser(BaseModel):
    """
    A class representing the instagram basic user info.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/user
    """
    account_type = attrib(default=None, type=Optional[str], repr=False)
    id = attrib(default=None, type=Optional[str])
    media_count = attrib(default=None, type=Optional[int], repr=False)
    username = attrib(default=None, type=Optional[str])


@attrs
class IgBasicMediaChildren(BaseModel):
    """
    A class representing the instagram basic media children media info.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/media/
    """
    id = attrib(default=None, type=Optional[str])
    media_type = attrib(default=None, type=Optional[str])
    media_url = attrib(default=None, type=Optional[str], repr=False)
    permalink = attrib(default=None, type=Optional[str])
    thumbnail_url = attrib(default=None, type=Optional[str], repr=False)
    timestamp = attrib(default=None, type=Optional[str], repr=False)
    username = attrib(default=None, type=Optional[str], repr=False)


@attrs
class IgBasicMedia(BaseModel):
    """
    A class representing the instagram basic media info.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/media/
    """
    caption = attrib(default=None, type=Optional[str], repr=False)
    id = attrib(default=None, type=Optional[str])
    media_type = attrib(default=None, type=Optional[str])
    media_url = attrib(default=None, type=Optional[str], repr=False)
    permalink = attrib(default=None, type=Optional[str])
    thumbnail_url = attrib(default=None, type=Optional[str], repr=False)
    timestamp = attrib(default=None, type=Optional[str], repr=False)
    username = attrib(default=None, type=Optional[str], repr=False)
    children = attrib(default=None, type=Optional[Dict], repr=False)

    def __attrs_post_init__(self):
        if self.children is not None and isinstance(self.children, dict):
            children = self.children.get("data", [])
            self.children = [IgBasicMediaChildren.new_from_json_dict(item) for item in children]



import json
import unittest

import responses

import pyfacebook


class ApiMediaTest(unittest.TestCase):
    BASE_PATH = "testdata/instagram/apidata/insights/"
    BASE_URL = "https://graph.facebook.com/{}/".format(pyfacebook.Api.VALID_API_VERSIONS[-1])

    with open(BASE_PATH + "user_insights.json", "rb") as f:
        USER_INSIGHTS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "user_insights_2.json", "rb") as f:
        USER_INSIGHTS_2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_album_insights.json", "rb") as f:
        MEDIA_ALBUM_INSIGHTS = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "media_photo_insights.json", "rb") as f:
        MEDIA_PHOTO_INSIGHTS = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.instagram_business_id = "123456"
        self.api = pyfacebook.IgProApi(
            app_id="123456", app_secret="secret",
            long_term_token="token"
        )

    def testGetUserInsights(self):
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/insights", json=self.USER_INSIGHTS)
            m.add("GET", self.BASE_URL + self.instagram_business_id + "/insights", json=self.USER_INSIGHTS_2)

            res = self.api.get_user_insights(
                user_id=self.instagram_business_id,
                period="day",
                metrics=["impressions", "follower_count"],
                since=1576771200,
                until=1577635200,
                access_token=self.api._access_token,
            )
            self.assertEqual(len(res), 2)
            self.assertEqual(res[0].name, "impressions")
            self.assertEqual(len(res[1].values), 10)

            res_json = self.api.get_user_insights(
                user_id=self.instagram_business_id,
                period="days_28",
                metrics=("impressions", "reach"),
                since=1576771200,
                until=1577635200,
                access_token=self.api._access_token,
                return_json=True,
            )
            self.assertEqual(len(res_json), 2)
            self.assertEqual(res_json[1]["name"], "reach")
            self.assertEqual(len(res_json[1]["values"]), 10)

    def testGetMediaInsights(self):
        media_album_id = "12345"
        media_photo_id = "67890"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + media_album_id + "/insights", json=self.MEDIA_ALBUM_INSIGHTS)
            m.add("GET", self.BASE_URL + media_photo_id + "/insights", json=self.MEDIA_PHOTO_INSIGHTS)

            media_album_insights = self.api.get_media_insights(
                media_id=media_album_id,
                metrics=[
                    "carousel_album_engagement", "carousel_album_impressions",
                    "carousel_album_reach", "saved", "video_views"
                ],
                access_token=self.api._access_token,
            )
            self.assertEqual(len(media_album_insights), 5)
            self.assertEqual(media_album_insights[0].period, "lifetime")
            self.assertEqual(media_album_insights[0].values[0].value, 138)

            media_photo_insights = self.api.get_media_insights(
                media_id=media_photo_id,
                metrics=[
                    "engagement", "impressions", "reach", "saved"
                ],
                access_token=self.api._access_token,
                return_json=True
            )
            self.assertEqual(len(media_photo_insights), 4)
            self.assertEqual(media_photo_insights[0]["values"][0]["value"], 65)



import json
import unittest

import responses
from six import iteritems

import pyfacebook


class CommentTestApi(unittest.TestCase):
    BASE_PATH = "testdata/facebook/apidata/comments/"
    BASE_URL = "https://graph.facebook.com/v8.0/"

    with open(BASE_PATH + "comment_by_parent_p1.json", "rb") as f:
        COMMENTS_PAGED_1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comment_by_parent_p2.json", "rb") as f:
        COMMENTS_PAGED_2 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "comment_info.json", "rb") as f:
        COMMENT_INFO = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "multi_default_fields.json", "rb") as f:
        MULTI_COMMENT_INFO_1 = json.loads(f.read().decode("utf-8"))
    with open(BASE_PATH + "multi_fields.json", "rb") as f:
        MULTI_COMMENT_INFO_2 = json.loads(f.read().decode("utf-8"))

    def setUp(self):
        self.api = pyfacebook.Api(
            app_id="123456", app_secret="secret",
            long_term_token="token", version="v8.0"
        )

    def testGetCommentByObject(self):
        post_id = "2121008874780932_2498613793687103"
        # test all comments
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + post_id + "/comments", json=self.COMMENTS_PAGED_1)
            m.add("GET", self.BASE_URL + post_id + "/comments", json=self.COMMENTS_PAGED_2)

            comments, comment_summary = self.api.get_comments_by_object(
                object_id=post_id,
                count=None,
                limit=5
            )

            self.assertEqual(len(comments), 7)
            self.assertEqual(comment_summary.total_count, 7)

        # test count
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + post_id + "/comments", json=self.COMMENTS_PAGED_1)

            comments, comment_summary = self.api.get_comments_by_object(
                object_id=post_id,
                count=5,
                limit=5,
                return_json=True
            )

            self.assertEqual(len(comments), 5)
            self.assertEqual(comment_summary["total_count"], 7)

    def testGetCommentInfo(self):
        comment_id = "2498598377021978_2498617433686739"
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL + comment_id, json=self.COMMENT_INFO)

            comment = self.api.get_comment_info(
                comment_id=comment_id,
            )
            self.assertEqual(comment.id, comment_id)

            comment_json = self.api.get_comment_info(
                comment_id=comment_id,
                return_json=True
            )
            self.assertEqual(comment_json["id"], comment_id)

    def testGetComments(self):
        ids = ["2498598377021978_2498617433686739", "2498598377021978_2509257692622713"]
        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL, json=self.MULTI_COMMENT_INFO_1)

            comment_dict = self.api.get_comments(
                ids=ids
            )
            for _id, data in iteritems(comment_dict):
                self.assertIn(_id, ids)
                self.assertEqual(_id, data.id)

        with responses.RequestsMock() as m:
            m.add("GET", self.BASE_URL, json=self.MULTI_COMMENT_INFO_2)
            comment_dict = self.api.get_comments(
                ids=ids,
                fields=["id", "message", "created_time", "comment_count"],
                return_json=True
            )
            for _id, data in iteritems(comment_dict):
                self.assertIn(_id, ids)
                self.assertEqual(_id, data["id"])



from .base import BaseApi
from .facebook import Api
from .instagram import IgProApi
from .instagram_basic import IgBasicApi


__all__ = [
    "BaseApi",
    "Api",
    "IgProApi",
    "IgBasicApi",
]
